# SUmmary experiment combinator
#
# Copyright (C) 2016--2022 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

import numpy                     # type: ignore
from epyc import ExperimentCombinator, Experiment, ResultsDict
import sys
from typing import List, Dict, Any
if sys.version_info >= (3, 8):
    from typing import Final
else:
    # backwards compatibility with Python 35, Python36, and Python37
    from typing_extensions import Final


class SummaryExperiment(ExperimentCombinator):
    """An experiment combinator that takes an underlying experiment and
    returns summary statistics for some of its results. This only really makes
    sense for experiments that return lists of results, such as those conducted
    using :class:`RepeatedExperiment`, but it works with any experiment.

    When run, a summary experiment summarises the experimental
    results, creating a new set of results that include the mean and
    variance for each result that the underyling experiments
    generated. (You can also select which results to summarise.) The
    raw results are discarded. The new results have the names of the
    raw results with suffices for mean, median, variance, and extrema.

    The summarisation obviously only works on result keys coming from the
    underlying experiments that are numeric. The default behaviour is to try to
    summarise all keys: you can restrict this by providing a list of keys to the
    constructor in the summarised_results keyword argument. Trying to summarise
    non-numeric results will be ignored (with a warining).

    The summary calculations only include those experimental runs that succeeded,
    that is that have their status set to True. Failed runs are ignored."""

    # Additional metadata
    UNDERLYING_RESULTS: Final[str] = 'epyc.summaryexperiment.repetitions'                          #: Metadata element for the number of results that were obtained.
    UNDERLYING_SUCCESSFUL_RESULTS: Final[str] = 'epyc.summaryexperiment.successful_repetitions'    #: Metadata elements for the number of results that were summarised.

    # Prefix and suffix tags attached to summarised result and metadata values
    MEAN_SUFFIX: Final[str] = '_mean'              #: Suffix for the mean of the underlying values.
    MEDIAN_SUFFIX: Final[str] = '_median'          #: Suffix for the median of the underlying values.
    VARIANCE_SUFFIX: Final[str] = '_variance'      #: Suffix for the variance of the underlying values.
    MIN_SUFFIX: Final[str] = '_min'                #: Suffix for the minimum of the underlying values.
    MAX_SUFFIX: Final[str] = '_max'                #: Suffix for the maximum of the underlying values.


    def __init__(self, ex: Experiment, summarised_results: List[str] = None):
        """Create a summarised version of the given experiment. The given
        fields in the experimental results will be summarised, defaulting to all.
        If there are fields that can't be summarised (because they're not
        numbers), remove them here.

        :param ex: the underlying experiment
        :param summarised_results: list of result values to summarise (defaults to all)"""
        super().__init__(ex)
        self._summarised_results = summarised_results

    def _mean(self, k: str) -> str:
        """Return the tag associated with the mean of k."""
        return k + self.MEAN_SUFFIX

    def _median(self, k: str) -> str:
        """Return the tag associated with the median of k."""
        return k + self.MEDIAN_SUFFIX

    def _variance(self, k: str) -> str:
        """Return the tag associated with the variance of k."""
        return k + self.VARIANCE_SUFFIX

    def _min(self, k: str) -> str:
        """Return the tag associated with the minimum of k."""
        return k + self.MIN_SUFFIX

    def _max(self, k: str) -> str:
        """Return the tag associated with the maximum of k."""
        return k + self.MAX_SUFFIX

    def summarise(self, results: List[ResultsDict]) -> ResultsDict:
        """Generate a summary of results from a list of experimental results dicts
        returned by running the underlying experiment. By default we generate
        mean, median, variance, and extrema for each value recorded.

        Override this method to create different or extra summary statistics.

        :param results: an array of experimental results dicts
        :returns: a dict of summary statistics"""
        if len(results) == 0:
            return {}
        else:
            summary = {}

            # work out the fields to summarise
            allKeys = results[0][Experiment.RESULTS].keys()
            ks = self._summarised_results
            if ks is None:
                # if we don't restrict, summarise all keys
                ks = allKeys
            else:
                # protect against a key that's not present
                ks = [k for k in ks if k in allKeys]

            # add the summary statistics
            for k in ks:
                # compute summaries for all fields we're interested in
                vs = [res[Experiment.RESULTS][k] for res in results]
                try:
                    summary[self._mean(k)]     = numpy.mean(vs)
                    summary[self._median(k)]   = numpy.median(vs)
                    summary[self._variance(k)] = numpy.var(vs)
                    summary[self._min(k)]      = numpy.min(vs)
                    summary[self._max(k)]      = numpy.max(vs)
                except Exception as e:
                    # couldn't do the statistics
                    print('Failed to summarise {k}: {e}'.format(k=k, e=e), file=sys.stderr)

            return summary

    def _flatten(self, rc: ResultsDict) -> List[ResultsDict]:
        '''Flatten-out any nested results.

        :param rc: a results dict, poissibly containing others nested as its results
        :returns: a flat list of results dicts'''
        rcs = []

        def _doflat(prc):
            if isinstance(prc[Experiment.RESULTS], list):
                for nrc in prc[Experiment.RESULTS]:
                    _doflat(nrc)
            else:
                rcs.append(prc)

        _doflat(rc)
        return rcs

    def do(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """Perform the underlying experiment and summarise its results.
        Our results are the summary statistics extracted from the results of
        the instances of the underlying experiment that we performed.

        We drop from the calculations any experiments whose completion status
        was False, indicating an error. Our own completion status will be
        True unless we had an error summarising a field (usually caused by trying
        to summarise non-numeric data).

        We record the exceptions generated by any experiment we summarise under
        the metadata key :attr:`SummaryExperiment.UNDERLYING_EXCEPTIONS`

        :param params: the parameters to the underlying experiment
        :returns: the summary statistics of the underlying results"""

        # perform the underlying experiment
        rc = self.experiment().run()

        # extract all the results as a single list
        rcs = self._flatten(rc)

        # extract only the successful runs
        sres = [rc for rc in rcs if rc[Experiment.METADATA][Experiment.STATUS]]

        # add extra values to our metadata record
        self._metadata[self.UNDERLYING_RESULTS]            = len(rcs)
        self._metadata[self.UNDERLYING_SUCCESSFUL_RESULTS] = len(sres)

        # construct summary results
        return self.summarise(sres)



# Base class for experiments
#
# Copyright (C) 2016--2022 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

import sys
from datetime import datetime
import traceback
import logging
from epyc import Logger
from typing import Set, Dict, Union, List, Any
if sys.version_info >= (3, 8):
    from typing import Final
else:
    # backwards compatibility with Python35, Python36, and Python37
    from typing_extensions import Final


logger = logging.getLogger(Logger)


# Type aliases
ResultsDict = Dict[str, Dict[str, Any]]     #: Type of results dicts.
ExperimentalParameters = Dict[str, Any]     #: Type of experimental parameter dicts.


class Experiment(object):
    """Base class for an :term:`experiment` conducted in a :term:`lab`.

    An :class:`Experiment` defines a computational experiment that can be run
    independently or (more usually) controlled from an instamnce of the
    :class:`Lab` class. Experiments should be long-lasting, able to conduct
    repeated runs at several different parameter points.

    From an experimenter's (or a lab's) perspective, an experiment
    has public methods :meth:`set` and :meth:`run`. The former sets the parameters
    for the experiment; the latter runs the experiment. producing a set
    of results that include direct experimental results and metadata on
    how the experiment ran. A single run may produce a list of result
    dicts if desired, each filled-in with the correct metadata.

    Experimental results, parameters, and metadata can be access directly
    from the :class:`Experiment` object. The class also exposes
    an indexing interface to access experimental results by name.
    """

    # Top-level structure for results dicts
    METADATA: Final[str] = 'metadata'                 #: Results dict key for metadata values, mainly timing.
    PARAMETERS: Final[str] = 'parameters'             #: Results dict key for describing the point in the parameter space the experiment ran on.
    RESULTS: Final[str] = 'results'                   #: Results dict key for the experimental results generated at the experiment's parameter point.

    # Standard metadata elements reported
    EXPERIMENT: Final[str] = 'epyc.experiment.classname'              #: Metadata element for storing the class name of the experiment.
    START_TIME: Final[str] = 'epyc.experiment.start_time'             #: Metadata element for the datetime experiment started.
    END_TIME: Final[str] = 'epyc.experiment.end_time'                 #: Metadata element for the datetime experiment ended.
    ELAPSED_TIME: Final[str] = 'epyc.experiment.elapsed_time'         #: Metadata element for the time experiment took overall in seconds.
    SETUP_TIME: Final[str] = 'epyc.experiment.setup_time'             #: Metadata element for the time spent on setup in seconds.
    EXPERIMENT_TIME: Final[str] = 'epyc.experiment.experiment_time'   #: Metadata element for the time spent on experiment itself in seconds.
    TEARDOWN_TIME: Final[str] = 'epyc.experiment.teardown_time'       #: Metadata element for the time spent on teardown in seconds.
    STATUS: Final[str] = 'epyc.experiment.status'                     #: Metadata element that will be True if experiment completed successfully, False otherwise.
    EXCEPTION: Final[str] = 'epyc.experiment.exception'               #: Metadata element containing the exception thrown if experiment failed.
    TRACEBACK: Final[str] = 'epyc.experiment.traceback'               #: Metadata element containing the traceback from the exception (as a string).

    # The above, collected together
    StandardMetadata: Set[str]                        #: The standard metadata elements to always capture.
    StandardMetadataTypes: Dict[str, type]            #: Type mapping for standard metadata.

    def __init__(self):
        self._metadata : Dict[str, Any] = dict()
        self._parameters : Dict[str, Any] = dict()
        self._results : Union[Dict[str, Any], List[Dict[str, Any]]] = dict()

    @classmethod
    def _init_statics(cls):
        '''Initialise the static members that need complex constructors.'''
        cls.StandardMetadata = set([ cls.EXPERIMENT,
                                     cls.START_TIME,
                                     cls.END_TIME,
                                     cls.ELAPSED_TIME,
                                     cls.EXPERIMENT_TIME,
                                     cls.SETUP_TIME,
                                     cls.TEARDOWN_TIME,
                                     cls.STATUS,
                                     cls.EXCEPTION,
                                     cls.TRACEBACK
                                   ])
        cls.StandardMetadataTypes = { cls.EXPERIMENT: str,
                                      cls.START_TIME: datetime,
                                      cls.END_TIME: datetime,
                                      cls.ELAPSED_TIME: float,
                                      cls.SETUP_TIME: float,
                                      cls.EXPERIMENT_TIME: float,
                                      cls.TEARDOWN_TIME: float,
                                      cls.STATUS: bool,
                                      cls.EXCEPTION: str,
                                      cls.TRACEBACK: str,
                                    }

    # ---------- Results dicts ----------

    @staticmethod
    def resultsdict() -> ResultsDict:
        '''Create an empty results dict, structured correctly.

        :returns: an empty results dict'''
        rc : ResultsDict = dict()
        rc[Experiment.PARAMETERS] = dict()
        rc[Experiment.METADATA] = dict()
        rc[Experiment.RESULTS] = dict()
        return rc


    # ---------- Configuration ----------

    def set(self, params: ExperimentalParameters) -> 'Experiment':
        """Set the parameters for the experiment, returning the
        now-configured experiment.

        :param params: the parameters
        :returns: the experiment"""
        if self._parameters is not None:
            self.deconfigure()
        self.configure(params)
        return self

    def configure(self, params: ExperimentalParameters):
        """Configure the experiment for the given parameters.
        The default stores the parameters for later use. Be sure
        to call this base method when overriding.

        :param params: the parameters"""
        self._parameters = params

    def deconfigure(self):
        """De-configure the experiment prior to setting new parameters.
        The default removes the parameters. Be sure
        to call this base method when overriding."""
        self._parameters = dict()


    # ---------- Running the experiment ----------

    def setUp(self, params: ExperimentalParameters):
        """Set up the experiment. Default does nothing.

        :param params: the parameters of the experiment"""
        pass

    def tearDown(self):
        """Tear down the experiment. Default does nothing."""
        pass

    def do(self, params: ExperimentalParameters) -> Union[Dict[str, Any], List[ResultsDict]]:
        """Do the body of the experiment. This should be overridden by
        sub-classes. Default does nothing.

        An experiment can return two types of results:

        - a dict mapping names to values for experimental results; or
        - a :term:`results dict` list, each of which represents a
          fully-formed experiment

        :param params: a dict of parameters for the experiment
        :returns: the experimental results

        """
        return dict()

    def report(self, params: ExperimentalParameters, meta: Dict[str, Any], res: Union[Dict[str, Any], List[ResultsDict]]) -> ResultsDict:
        """Return a properly-structured dict of results. The default returns a
        dict with results keyed by :attr:`Experiment.RESULTS`, the
        data point in the parameter space keyed by
        :attr:`Experiment.PARAMETERS`, and timing and other metadata
        keyed by :attr:`Experiment.METADATA`. Overriding this method
        can be used to record extra metadata values, but be sure to
        call the base method as well.

        If the experimental results are a list of results dicts, then
        we report a results dict whose results are a list of results
        dicts.  This is used by :class:`RepeatedExperiment` amd other
        experiments that want to report multiple sets of results.

        :param params: the parameters we ran under
        :param meta: the metadata for this run
        :param res: the direct experimental results from do()
        :returns: a :term:`results dict`

        """
        rc = Experiment.resultsdict()
        rc[self.PARAMETERS] = params.copy()
        rc[self.METADATA] = meta.copy()
        rc[self.RESULTS] = res
        return rc

    def run(self, fatal: bool = False) -> ResultsDict:
        """Run the experiment, using the parameters set using :meth:`set`.  A
        "run" consists of calling :meth:`setUp`, :meth:`do`, and
        :meth:`tearDown`, followed by collecting and storing (and
        returning) the experiment's results. If running the experiment
        raises an exception, that will be returned in the metadata
        along with its traceback to help with experiment debugging.
        If fatal is True it will also be raised from this method: the
        default prints the exception but doesn't raise it.

        :param fatal: (optional) raise any exceptions (default suppresses them into the results dict)
        :returns: a :term:`results dict`

        """

        # record the experiment's class name
        self._metadata = dict()
        cn = '{modulename}.{classname}'.format(modulename=self.__class__.__module__,
                                               classname=self.__class__.__name__)
        self._metadata[self.EXPERIMENT] = cn

        # perform the experiment protocol
        params = self.parameters()
        self._results = dict()
        res = dict()
        doneSetupTime = doneExperimentTime = doneTeardownTime = None
        elapsedTime = 0
        startTime = datetime.now()
        self._metadata[self.START_TIME] = startTime
        try:
            # do the phases in order, recording the wallclock times at each phase
            self.setUp(params)
            doneSetupTime = datetime.now()
            dt = (doneSetupTime - startTime).total_seconds()
            elapsedTime = dt
            self._metadata[self.SETUP_TIME] = dt
            res = self.do(params)
            doneExperimentTime = datetime.now()
            dt = (doneExperimentTime - doneSetupTime).total_seconds()
            elapsedTime += dt
            self._metadata[self.EXPERIMENT_TIME] = dt
            self.tearDown()
            doneTeardownTime = datetime.now()
            dt = (doneTeardownTime - doneExperimentTime).total_seconds()
            elapsedTime += dt
            self._metadata[self.TEARDOWN_TIME] = dt
            self._metadata[self.END_TIME] = doneTeardownTime
            self._metadata[self.ELAPSED_TIME] = elapsedTime

            # set the success flag
            self._metadata[self.STATUS] = True
        except Exception as e:
            # grab the traceback before we do anything else
            #_, _, tb = sys.exc_info()
            tb = traceback.format_exc()

            # decide on the cleanup actions that need doing
            if (doneSetupTime is not None) and (doneExperimentTime is None):
                # we did the setup and then failed in the experiment, so
                # we need to do the teardown
                try:
                    self.tearDown()
                except Exception as f:
                    # log, but otherwise ignore, any exceptions
                    # that happen in the teardown
                    logger.error(f'Caught exception in teardown (ignored): {f}')
            self._metadata[self.ELAPSED_TIME] = elapsedTime
            self._metadata[self.END_TIME] = datetime.now()

            # set the failure flag and record the exception
            self._metadata[self.STATUS] = False
            self._metadata[self.EXCEPTION] = e
            self._metadata[self.TRACEBACK] = tb

            # propagate the exception or just report it
            if fatal:
                raise e
            else:
                logger.error(f'Caught exception in experiment: {e}')

        # report the results
        self._results = res
        return self.report(params,
                           self._metadata,
                           res)


    # ---------- Accessing results ----------

    def __getitem__(self, k: str) -> Any:
        """Return the given element of the experimental results. This only
        gives access to the experimental results, *not* to the
        parameters or metadata.

        :param k: the result key
        :returns: the value
        :raises: KeyError if there is no such result

        """
        return (self.experimentalResults())[k]

    def success(self) -> bool:
        """Test whether the experiment has been run successfully. This will be
        False if the experiment hasn't been run, or if it's been run
        and failed.

        :returns: True if the experiment has been run successfully

        """
        if self.STATUS in self.metadata().keys():
            return (self.metadata())[self.STATUS]
        else:
            return False

    def failed(self) -> bool:
        '''Test whether an experiment failed. This will be True if the
        experiment has been run and has failed, which means that there
        will be an exception and traceback information stored in the
        metadata. It will be False if the experiment hasn't been run.

        :returns: True if the experiment has failed

        '''
        if self.STATUS not in self.metadata().keys():
            return False
        else:
            return not (self.metadata())[self.STATUS]

    def results(self) -> Union[ResultsDict, List[ResultsDict]]:
        """Return a complete results dict. Only really makes sense for
        recently-executed experimental runs.

        :returns: the results dict, or a list of them"""
        return self.report(self.parameters(),
                           self.metadata(),
                           self.experimentalResults())

    def experimentalResults(self) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        """Return the experimental results from our last run. This will be
        None if we haven't been run, or if we ran and failed.

        :returns: the experimental results dict, which may be empty, and may be a list of dicts

        """
        return self._results

    def parameters(self) -> Dict[str, Any]:
        """Return the current experimental parameters, which will be None if
        none have been given by a call to set()

        :returns: the parameters, which may be empty

        """
        return self._parameters

    def metadata(self) -> Dict[str, Any]:
        """Return the metadata we collected at out last execution, which will
        be None if we've not been executed and an empty dict if we're
        mid-run (i.e., if this method is called from do() for whatever
        reason).

        :returns: the metadata, which may be empty

        """
        return self._metadata



# Simulation "lab" experiment management, sequential version
#
# Copyright (C) 2016--2022 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

import logging
from epyc import Logger, LabNotebook, Experiment, Design, FactorialDesign, ResultsDict, ExperimentalConfiguration
from pandas import DataFrame                                   # type: ignore
from typing import Dict, List, Tuple, Any, Callable


logger = logging.getLogger(Logger)


class Lab(object):
    """A laboratory for computational experiments.

    A :class:`Lab` conducts an experiment at different points in a
    multi-dimensional parameter space.  The default performs all the
    experiments locally; sub-classes exist to perform remote parallel
    experiments.

    A :class:`Lab` stores its result in a notebook, an instance of
    :class:`LabNotebook`.  By default the base :class:`Lab` class uses
    an in-memory notebook, essentially just a dict; sub-classes use
    persistent notebooks to manage larger sets of experiments.

    Each lab has an associated :class:`Design` that turns a set of
    parameter ranges into a set of individual "points" of the parameter
    space at which to perform actual experiments. The default is to
    use a :class:`FactorialDesign` that performs an experiment for every
    combination of parameter values. This might be a *lot* of experiments,
    and other designs can be used to reduce or modify the space.

    :param notebook: the notebook used to store results (defaults to an empty :class:`LabNotebook`)
    :param design: the experimental design to use (defaults to a :class:`FactorialDesign`)
    """

    def __init__(self, notebook: LabNotebook = None, design: Design = None):
        self._notebook = notebook if notebook is not None else LabNotebook()
        self._design = design if design is not None else FactorialDesign()
        self._parameters : Dict[str, Any] = dict()

    def notebook(self) -> LabNotebook:
        """Return the notebook being used by this lab.

        :returns: the notebook"""
        return self._notebook

    def design(self) -> Design:
        '''Return the experimental design this lab uses.

        :returns: the design'''
        return self._design


    # ---------- Protocol ----------

    def open( self ):
        """Open a lab for business. Sub-classes might insist the they are
        opened and closed explicitly when experiments are being performed.
        The default does nothing."""
        pass

    def close( self ):
        """Shut down a lab. Sub-classes might insist the they are
        opened and closed explicitly when experiments are being performed.
        The default does nothing."""
        pass

    def updateResults( self ):
        """Update the lab's results. This method is called by all other methods
        that return results in some sense, and may be overridden to let the results
        "catch up" with external processing. The default does nothing."""
        pass

    def recreate(self):
        '''Return a structure describing this lab in enough detail to reconstruct it,
        consisting of the name of the class and a dict of any arguments that it needs.
        Sub-classes should call the base method to fill in any defaults and then add
        any arguments they need to the dict.

        :returns: a (classname, args) pair
        '''
        n = '{modulename}.{classname}'.format(modulename = self.__class__.__module__,
                                              classname = self.__class__.__name__)
        args = dict()
        return (n, args)


    # ---------- Managing experimental parameters ----------

    def addParameter(self, k : str, r : Any):
        """Add a parameter to the experiment's parameter space. k is the
        parameter name, and r is its range. The range can be a single value
        or a list, or any other iterable. (Strings are counted as single values.)

        :param k: parameter name
        :param r: parameter range"""
        if isinstance(r, str):
            # strings are single values
            self._parameters[k] = [ r ]
        else:
            try:
                # try to unpack using iterator
                self._parameters[k] = list(r)
            except TypeError:
                # not iterable, a single value
                self._parameters[k] = [ r ]

    def deleteParameter(self, k : str):
        '''Delete a parameter from the parameter space. If the
        parameter doesn't exist then this is a no-op.

        :param k: the parameter name'''
        if k in self._parameters:
            del self._parameters[k]

    def deleteAllParameters(self):
        '''Delete all parameters from the parameter space.'''
        for k in self.parameters():
            del self[k]

    def parameters(self) -> List[str]:
        """Return a list of parameter names.

        :returns: a list of parameter names"""
        return list(self._parameters.keys())

    def __len__(self) -> int:
        """The length of an experiment is the total number of data points
        that will be explored. This is the length of the experimental
        configuration returned by :meth:`experiments`.

        :returns: the number of experimental runs"""
        return len(self.experiments(None))

    def __getitem__(self, k : str) -> Any:
        """Access a parameter range using array notation.

        :param k: parameter name
        :returns: the parameter range"""
        return self._parameters[k]

    def __setitem__(self, k : str, r : Any):
        """Add a parameter using array notation.

        :param k: the parameter name
        :param r: the parameter range"""
        self.addParameter(k, r)

    def __delitem__(self, k : str):
        '''Delete a parameter using array notation.

        :param k: the key'''
        self.deleteParameter(k)

    def __contains__(self, k : str) -> bool:
        '''Test whether the given key is a paramater for the lab.

        :param k: the keys
        :returns: True if this is a parameter'''
        return k in self.parameters()

    def experiments(self, e: Experiment) -> ExperimentalConfiguration:
        """Return the :term:`experimental configuration`, a list consisting of
        experiments and the points at which they should be run. The
        structure of the experimental space is defined by the lab's
        experimental design, which may also change the experiment to
        be run.

        :param e: the experiment
        :returns: an experimental configuration

        """
        return self._design.experiments(e, self._parameters)


    # ---------- Running experiments ----------

    def runExperiment(self, e: Experiment):
        """Run an experiment over all the points in the parameter space.
        The results will be stored in the notebook.

        :param e: the experiment"""

        # create the experimental parameter space
        eps = self.experiments(e)

        # run the experiment at each point
        nb = self.notebook()
        for (ep, p) in eps:
            res = ep.set(p).run()
            nb.addResult(res)

        # commit the results
        nb.commit()


    # ---------- Conditional result set creation ----------

    def createWith(self, tag: str, f: Callable[['Lab'], bool],
                   description: str = None,
                   propagate: bool = True, delete: bool = True,
                   finish: bool = False,
                   deleteAllParameters: bool = True):
        '''Use a function to create a result set.

        If the result set already exists in the lab's notebook, it is
        selected; if it doesn't, it is created, selected, and the
        creation function is called. The creation function is passed
        a reference to the lab it is populating.

        By default any exception in the creation function will cause
        the incomplete result set to be deleted and the previously
        current result set to be re-selected: this can be inhibited by
        setting ``delete=False``. Any raised exception is propagated by
        default: this can be inhibited by setting ``propagate =
        False``. The result set can be locked after creation by
        setting ``finished=True``, as long as the creation was successful:
        poorly-created result sets aren't locked.

        By default the lab has its parameters cleared before calling the
        creation function, so that it occurd "clean". Set
        ``deleteAllParameters=False`` to inhibit this.

        :param tag: the result set tag
        :param f: the creation function (taking Lab as argument)
        :param description: (optional) description if a result set is created
        :param propagate: (optional) propagate any excepton (defaults to True)
        :param delete: (optional) delete on exception (default is True)
        :param finish: (optional) lock the result set after creation (defaults to False)
        :param deleteAllParameters: (optional) delete all lab parameters before creation (defaults to True)
        :returns: True if the result set exists already or was properly created

        '''
        nb = self.notebook()

        # grab the tag of the current result set
        ctag = nb.currentTag()

        # select and return if the result set already exists
        if nb.already(tag, description):
            return True

        # if we get here, the result set will have been be created and selected
        try:
            if deleteAllParameters:
                # clear the lab's parameter space ahead of creation
                self.deleteAllParameters()

            # call the creation function
            f(self)

            # lock the result set if requested
            if finish:
                nb.current().finish()

            # if we get here, we were successful
            return True
        except Exception as e:
            # if we get here, creation failed
            if delete:
                # re-select the previous current result set
                try:
                    nb.select(ctag)
                    logger.info(f'Reverted to ResultSet {ctag} in error')
                except Exception:
                    # the creation function messed with the results sets,
                    # so revert to the default (which is always present)
                    nb.select(LabNotebook.DEFAULT_RESULTSET)
                    logger.warning('Reverted to ResultSet {dtag} ({ctag} was deleted)'.format(ctag=ctag,
                                                                                              dtag=LabNotebook.DEFAULT_RESULTSET))

                # delete the partial set
                nb.deleteResultSet(tag)
                logger.info(f'Deleted partly-created result set {tag}')

            if propagate:
                # propagate the exception
                raise e
            else:
                logger.error(f'Exception ignored in createWith(): {e}')
                return False


    # ---------- Accessing results ----------

    def dataframe(self, only_successful : bool =True) -> DataFrame:
        """Return the current results as a pandas DataFrame after resolving
        any pending results that have completed. This makes
        use of the underlying notebook's current result set. For finer control,
        access the notebook's :meth:`LabNotebook.dataframe` or
        :meth:LabNotebook.dataframeFor` methods directly.

        :param only_successful: only return successful results
        :returns: the resulting dataset as a DataFrame"""
        self.updateResults()
        return self._notebook.dataframe(only_successful=only_successful)

    def results(self) -> List[ResultsDict]:
        '''Return the current results as a list of results dicts after resolving
        any pending results that have completed. This makes
        use of the underlying notebook's current result set. For finer control,
        access the notebook's :meth:`LabNotebook.results` or
        :meth:LabNotebook.resultsFor` methods directly.

        Note that this approach to acquiring results is a lot slower and more
        memory-hungry than using :meth:`dataframe`, but may be useful for small
        sets of results that benefit from a more Pythonic intertface.'''
        self.updateResults()
        return self._notebook.results()

    def ready(self, tag : str =None) -> bool:
        """Test whether all the results are ready in the tagged result
        set -- that is, none are pending.

        :param tag: (optional) the result set to check (default is the current result set)
        :returns: True if the results are in"""
        self.updateResults()
        return self._notebook.ready(tag)

    def readyFraction(self, tag : str =None) -> float:
        '''Return the fraction of results available (not pending) in the
        tagged result set after first updating the results.

        :param tag: (optional) the result set to check (default is the current result set)
        :returns: the ready fraction'''
        self.updateResults()
        return self._notebook.readyFraction(tag)



# Tests of HDF5 notebooks
#
# Copyright (C) 2020 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import *
import unittest
import numpy
import os
from tempfile import NamedTemporaryFile


# Remote HDF5 file for testiong URL behaviour
testFileURL = 'https://raw.githubusercontent.com/simoninireland/epyc/dev/test/test.h5'


class SampleExperiment(Experiment):
    '''A very simple experiment that adds up its parameters.'''

    def do( self, param ):
        return dict(total=param['k'] + 10)


class SampleExperiment1(Experiment):
    '''An experiment that fails.'''

    def do( self, param ):
        raise Exception('A (deliberate) failure')


class SampleExperiment2(Experiment):
    '''An experiment whose results contain a list.'''

    def do( self, param ):
        k = param['k']
        return dict(list=[k] * k)


class HDF5LabNotebookTests(unittest.TestCase):

    def setUp( self ):
        '''Set up with a temporary file.'''
        tf = NamedTemporaryFile()
        tf.close()
        self._fn = tf.name
        #self._fn = 'test.h5'

    def tearDown( self ):
        '''Delete the temporary file.'''
        try:
            os.remove(self._fn)
            #pass
        except OSError:
            pass

    def testCreate( self ):
        '''Test creation of empty notebook (which will create a backing file too)'''
        nb = HDF5LabNotebook(self._fn)
        self.assertTrue(nb.isPersistent())
        self.assertEqual(nb.name(), self._fn)
        self.assertTrue(os.path.isfile(self._fn))

    def testTypeDescription(self):
        '''Test we can extract the type description of an experiment.'''
        nb = HDF5LabNotebook(self._fn, create=True)
        params = dict()
        params['k'] = 3
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)

    def testReadWrite(self):
        '''Test we can read and write results.'''
        nb = HDF5LabNotebook(self._fn, create=True)
        params = dict()
        params['k'] = 3
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        nb.commit()

        # check we read back the same results
        nb = HDF5LabNotebook(self._fn)
        res = nb.dataframe().iloc[0]
        for d in [ Experiment.METADATA, Experiment.PARAMETERS, Experiment.RESULTS ]:
            for k in rc[d].keys():
                self.assertEqual(res[k], rc[d][k])

        # check we reconstruct the results dict form correctly too
        # (indirectly checks writing and reading dataset attributes)
        rc1 = nb.results()[0]
        for d in [ Experiment.METADATA, Experiment.PARAMETERS, Experiment.RESULTS ]:
            for k in rc1[d].keys():
                self.assertEqual(res[k], rc1[d][k])

    def testReadWritePending(self):
        '''Test we can read and write result sets with pending results.'''
        nb = HDF5LabNotebook(self._fn, create=True)
        params = dict()
        params['k'] = 3
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        params['k'] = 5
        nb.addPendingResult(params, '1234')
        nb.commit()

        nb = HDF5LabNotebook(self._fn)
        self.assertEqual(nb.numberOfPendingResults(), 1)
        self.assertCountEqual(nb.current().pendingResults(), [ '1234' ])

    def testReadWriteAllPending(self):
        '''Test reading and writing a result set with just pending results,'''
        nb = HDF5LabNotebook(self._fn, create=True)
        params = dict()
        params['k'] = 5
        nb.addPendingResult(params, '1234')
        params['k'] = 8
        nb.addPendingResult(params, '4567')
        nb.commit()

        nb = HDF5LabNotebook(self._fn)
        self.assertEqual(nb.numberOfPendingResults(), 2)
        self.assertCountEqual(nb.current().pendingResults(), [ '1234', '4567' ])
        self.assertEqual(nb.numberOfResults(), 0)

    def testResolveAcrossReadWrite(self):
        '''Test we can resolve a pending result across a read/write cycle.'''
        nb = HDF5LabNotebook(self._fn, create=True)

        # set up two pending results and save them
        params = dict()
        params['k'] = 5
        nb.addPendingResult(params, '1234')
        params['k'] = 8
        nb.addPendingResult(params, '4567')
        self.assertCountEqual(nb.current().pendingResults(), [ '1234', '4567' ])
        nb.commit()

        # re-load the notebook and resolve one of the results, saving again
        nb = HDF5LabNotebook(self._fn)
        params['k'] = 8
        rc1 = SampleExperiment().set(params).run()
        nb.resolvePendingResult(rc1, '4567')
        self.assertEqual(nb.numberOfPendingResults(), 1)
        self.assertCountEqual(nb.current().pendingResults(), [ '1234' ])
        self.assertEqual(nb.numberOfResults(), 1)
        rc1p = nb.results()[0]
        for d in [ Experiment.PARAMETERS, Experiment.RESULTS ]:
            for k in rc1p[d].keys():
                self.assertEqual(rc1p[d][k], rc1[d][k])
        nb.commit()

        # check the next save went OK
        nb = HDF5LabNotebook(self._fn)
        self.assertEqual(nb.numberOfPendingResults(), 1)
        self.assertEqual(nb.numberOfResults(), 1)
        rc1p = nb.results()[0]
        for d in [ Experiment.PARAMETERS, Experiment.RESULTS ]:
            for k in rc1p[d].keys():
                self.assertEqual(rc1p[d][k], rc1[d][k])

        # resolve the second result, check their integrity, and save again
        nb = HDF5LabNotebook(self._fn)
        params['k'] = 5
        rc2 = SampleExperiment().set(params).run()
        nb.resolvePendingResult(rc2, '1234')
        self.assertEqual(nb.numberOfPendingResults(), 0)
        self.assertEqual(nb.numberOfResults(), 2)
        res = nb.results()
        if res[0][Experiment.PARAMETERS]['k'] == 5:
            ts = [ (res[0], rc2), (res[1], rc1) ]
        else:
            ts = [ (res[0], rc1), (res[1], rc2) ]
        for (resp, rcp) in ts:
            for d in [ Experiment.PARAMETERS, Experiment.RESULTS ]:
                for k in rcp[d].keys():
                    self.assertEqual(rcp[d][k], resp[d][k])
        nb.commit()

        # final reload and sanity check
        nb = HDF5LabNotebook(self._fn)
        self.assertEqual(nb.numberOfPendingResults(), 0)
        self.assertEqual(nb.numberOfResults(), 2)

    def testReadWriteEmpty(self):
        '''Test an empty notebook is still empty after saving.'''
        nb = HDF5LabNotebook(self._fn, create=True)
        nb.commit()

        nb = HDF5LabNotebook(self._fn)
        self.assertEqual(nb.numberOfResultSets(), 1)
        self.assertEqual(nb.numberOfResults(), 0)
        self.assertEqual(nb.numberOfPendingResults(), 0)

    def testAddResultToEmpty(self):
        '''Test we can add to an empty notebook after saving.'''
        nb = HDF5LabNotebook(self._fn, create=True)
        nb.commit()

        nb = HDF5LabNotebook(self._fn)
        params = dict()
        params['k'] = 3
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        self.assertEqual(nb.numberOfResults(), 1)
        self.assertEqual(nb.numberOfPendingResults(), 0)

    def testAddPendingResultToEmpty(self):
        '''Test we can add a pending result to an empty notebook after saving.'''
        nb = HDF5LabNotebook(self._fn, create=True)
        nb.commit()

        nb = HDF5LabNotebook(self._fn)
        params = dict()
        params['k'] = 3
        nb.addPendingResult(params, '1234')
        self.assertEqual(nb.numberOfResults(), 0)
        self.assertEqual(nb.numberOfPendingResults(), 1)
        nb.commit()

        nb = HDF5LabNotebook(self._fn)
        self.assertEqual(nb.numberOfResults(), 0)
        self.assertEqual(nb.numberOfPendingResults(), 1)

    def testFailure(self):
        '''Test we can add successful and failed results.'''
        nb = HDF5LabNotebook(self._fn, create=True)

        # add the successful result
        params = dict()
        params['k'] = 3
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        nb.commit()
        t = rc[Experiment.RESULTS]['total']   # grab for later test

        # add the failed result
        rc[Experiment.METADATA][Experiment.STATUS] = False
        rc[Experiment.METADATA][Experiment.EXCEPTION] = 'An exception'
        rc[Experiment.METADATA][Experiment.TRACEBACK] = 'Called from somewhere'
        rc[Experiment.RESULTS] = dict()
        nb.addResult(rc)
        nb.commit()

        # check the results were stored correctly
        nb = HDF5LabNotebook(self._fn)
        for rc in nb.results():
            if rc[Experiment.METADATA][Experiment.STATUS]:
                # the successful result, check we have all result fields
                self.assertCountEqual(rc[Experiment.RESULTS].keys(), ['total'])
                self.assertEqual(rc[Experiment.RESULTS]['total'], t)

                # make sure the missing exception and traceback fields were properly zeroed
                self.assertEqual(rc[Experiment.METADATA][Experiment.EXCEPTION], '')
                self.assertEqual(rc[Experiment.METADATA][Experiment.TRACEBACK], '')
            else:
                # failed result, shouldn't be any result fields
                self.assertCountEqual(rc[Experiment.RESULTS].keys(), [])

    def testMaintainType(self):
        '''Test we can add several results without rebuilding the file unnecessarily.'''
        nb = HDF5LabNotebook(self._fn, create=True)

        # add two results, same type
        params = dict()
        params['k'] = 3
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        self.assertTrue(nb.current().isTypeChanged())
        params['k'] = 4
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        self.assertTrue(nb.current().isTypeChanged())

        # check commitment clears the flags
        nb.commit()
        self.assertFalse(nb.current().isTypeChanged())
        self.assertFalse(nb.current().isDirty())

        # add another value and check the type isn't changed
        params['k'] = 5
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        self.assertFalse(nb.current().isTypeChanged())

    def testExtraMetadata(self):
        '''Test we can add and save extra metadata fields.'''
        nb = HDF5LabNotebook(self._fn, create=True)

        # create a result and add some extra metadata
        params = dict()
        params['k'] = 3
        rc = SampleExperiment().set(params).run()
        rc[Experiment.METADATA]['extra'] = 'some more'
        nb.addResult(rc)
        nb.commit()

        # check the extra field comes back in
        nb = HDF5LabNotebook(self._fn)
        rc1 = nb.results()[0]
        self.assertEqual(rc1[Experiment.METADATA]['extra'], rc[Experiment.METADATA]['extra'])

    def testTwoResultSets(self):
        '''Test we can maintain two result sets without interference.'''
        nb = HDF5LabNotebook(self._fn, create=True)

        # create a couple of results and pending results in the first result set
        params = dict()
        params['k'] = 3
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        params['k'] = 4
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        params['k'] = 5
        nb.addPendingResult(params, '1234')
        nb.commit()

        # create more results with a different type in a new result set
        nb.addResultSet('second')
        params['k'] = 40
        params['s'] = 60
        rc = SampleExperiment().set(params).run()
        rc[Experiment.METADATA]['extra'] = 'some more'
        nb.addResult(rc)
        params['k'] = 50
        params['s'] = 70
        rc = SampleExperiment().set(params).run()
        rc[Experiment.METADATA]['extra'] = 'some more'
        nb.addResult(rc)
        nb.commit()

        # check the file has the same structure
        nb = HDF5LabNotebook(self._fn)
        self.assertCountEqual(nb.resultSets(), [ LabNotebook.DEFAULT_RESULTSET, 'second' ])
        self.assertEqual(nb.currentTag(), 'second')   # the current set when we committed
        nb.select(LabNotebook.DEFAULT_RESULTSET)
        self.assertEqual(nb.current().numberOfResults(), 2)
        self.assertEqual(nb.current().numberOfPendingResults(), 1)
        res = nb.results()[0]
        res = nb.results()[0]
        self.assertIn('k', res[Experiment.PARAMETERS])
        self.assertNotIn('s', res[Experiment.PARAMETERS])
        self.assertNotIn('extra', res[Experiment.METADATA])
        nb.select('second')
        self.assertEqual(nb.current().numberOfResults(), 2)
        self.assertEqual(nb.current().numberOfPendingResults(), 0)
        res = nb.results()[0]
        self.assertIn('k', res[Experiment.PARAMETERS])
        self.assertIn('s', res[Experiment.PARAMETERS])
        self.assertIn('extra', res[Experiment.METADATA])

    def testCurrentWithoutChange(self):
        '''Test we record a change of current result set with no other action.'''
        nb = HDF5LabNotebook(self._fn, create=True)

        # create a couple of results sets
        params = dict()
        params['k'] = 3
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        params['k'] = 4
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        nb.addResultSet('second')
        params['k'] = 40
        params['s'] = 60
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        params['k'] = 50
        params['s'] = 70
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        nb.commit()

        # check the current set
        nb = HDF5LabNotebook(self._fn)
        self.assertCountEqual(nb.resultSets(), [ LabNotebook.DEFAULT_RESULTSET, 'second' ])
        self.assertEqual(nb.currentTag(), 'second')

        # change the current set, save, re-load, and check
        nb.select(LabNotebook.DEFAULT_RESULTSET)
        nb.commit()
        nb = HDF5LabNotebook(self._fn)
        self.assertEqual(nb.currentTag(), LabNotebook.DEFAULT_RESULTSET)

    def testInferList(self):
        '''Test we can add results that contain lists.'''
        nb = HDF5LabNotebook(self._fn, create=True)

        # add a result with a list
        params = dict()
        params['k'] = 3
        rc = SampleExperiment2().set(params).run()
        nb.addResult(rc)

        # add another with the same shape1
        params['k'] = 3
        rc = SampleExperiment2().set(params).run()
        nb.addResult(rc)
        nb.commit()

        # check the values have the right shapes
        nb = HDF5LabNotebook(self._fn)
        params['k'] = 3
        rc = nb.resultsFor(params)[0]
        self.assertCountEqual(rc[Experiment.RESULTS]['list'], [ 3, 3, 3 ])

    def testContextManager(self):
        '''Test that the conext manager works as inteneded.'''
        nb = HDF5LabNotebook(self._fn, create=True)
        with nb.open():
            # add two results, same type
            params = dict()
            params['k'] = 3
            rc = SampleExperiment().set(params).run()
            nb.addResult(rc)
            params['k'] = 4
            rc = SampleExperiment().set(params).run()
            nb.addResult(rc)

        # re-load and check we have the last result
        nb = HDF5LabNotebook(self._fn)
        self.assertEqual(nb.numberOfResults(), 2)
        self.assertEqual(len(nb.dataframeFor(params)), 1)

    def testContextManagerExceptions(self):
        '''Test that the context manager is robust to exceptions.'''
        nb = HDF5LabNotebook(self._fn, create=True)
        try:
            with nb.open():
                # add two results, same type -- but jump out before the second one gets added
                params = dict()
                params['k'] = 3
                rc = SampleExperiment().set(params).run()
                nb.addResult(rc)
                params['k'] = 4
                raise Exception('second one is skipped')

                # this is what we then shouldn't see, because of the exception
                rc = SampleExperiment().set(params).run()
                nb.addResult(rc)
        except:
            pass

        # re-load and check we have the first result (and not the second)
        nb = HDF5LabNotebook(self._fn)
        self.assertEqual(nb.numberOfResults(), 1)
        params = dict()
        params['k'] = 3
        self.assertEqual(len(nb.dataframeFor(params)), 1)
        params['k'] = 4
        self.assertEqual(len(nb.dataframeFor(params)), 0)

    def testAttributes(self):
        '''Test we can read and write attributes.'''
        nb = HDF5LabNotebook(self._fn, description='A test notebook', create=True)

        # attributes of a result set but no results dataset
        rs = nb.current()
        rs['number1'] = '1'
        rs['number2'] = 2
        nb.commit()
        nb = HDF5LabNotebook(self._fn)
        rs = nb.current()
        self.assertEqual(rs['number1'], '1')
        self.assertEqual(rs['number2'], '2')    # all attributes stored as strings at present

        # check we pulled the description as well
        self.assertEqual(nb.description(), 'A test notebook')

        # add a new result set and some results, check attributes
        nb.addResultSet('new', 'Another result set')
        rs = nb.current()
        rs['number1'] = '4'
        rs['number3'] = '3'
        params = dict()
        params['k'] = 3
        rc = SampleExperiment().set(params).run()
        nb.addResult(rc)
        nb.commit()
        nb = HDF5LabNotebook(self._fn)
        rs = nb.current()
        self.assertEqual(rs.description(), 'Another result set')
        self.assertEqual(len(rs), 1)
        self.assertEqual(rs['number3'], '3')
        self.assertEqual(rs['number1'], '4')
        rs = nb.select(LabNotebook.DEFAULT_RESULTSET)
        self.assertEqual(rs['number1'], '1')
        self.assertEqual(rs['number2'], '2')    # all attributes stored as strings at present

        # test that deleting attributes from the result set deletes them from the file
        del rs['number1']
        nb.commit()
        nb = HDF5LabNotebook(self._fn)
        rs = nb.current()
        self.assertCountEqual(rs.keys(), ['number2'])

    def testOverrideDescription(self):
        '''Test that the description can be overrideen on a loaded file.'''
        with HDF5LabNotebook(self._fn, description='A test notebook', create=True).open() as nb:
            pass

        with HDF5LabNotebook(self._fn, description='A new description').open() as nb:
            self.assertEqual(nb.description(), 'A new description')

    def testChangeAttributes(self):
        '''Test we can change attributes, descriptions, etc.'''
        with HDF5LabNotebook(self._fn, description='A test notebook', create=True).open() as nb:
            pass

        # check the description matches
        with HDF5LabNotebook(self._fn).open() as nb:
            self.assertEqual(nb.description(), 'A test notebook')

            # check we can change the description
            nb.setDescription('A new description')
        with HDF5LabNotebook(self._fn).open() as nb:
            self.assertEqual(nb.description(), 'A new description')

        # check we can set and change a result set description
        with HDF5LabNotebook(self._fn).open() as nb:
            nb.addResultSet('second', 'A result set')
        with HDF5LabNotebook(self._fn).open() as nb:
            rs = nb.current()
            self.assertEqual(rs.description(), 'A result set')
            rs.setDescription('Changed')
        with HDF5LabNotebook(self._fn).open() as nb:
            self.assertEqual(rs.description(), 'Changed')

        # check we can set, change, and delete result set attributes
        with HDF5LabNotebook(self._fn).open() as nb:
            rs = nb.current()
            rs['name'] = 'an attribute'
            rs['date'] = 'today'
        with HDF5LabNotebook(self._fn).open() as nb:
            rs = nb.current()
            self.assertEqual(rs['name'], 'an attribute')
            self.assertEqual(rs['date'], 'today')
            rs['date'] = 'tomorrow'
            del rs['name']
        with HDF5LabNotebook(self._fn).open() as nb:
            rs = nb.current()
            self.assertCountEqual(rs.keys(), [ 'date' ])
            self.assertEqual(rs['date'], 'tomorrow')

    def testNoDefault(self):
        '''Test things work when the default resultset is left empty.'''

        # get some results for later
        params = dict()
        params['k'] = 3
        rc = SampleExperiment().set(params).run()

        # put a pending result into a new dataset, then commit
        with HDF5LabNotebook(self._fn, create=True).open() as nb:
            nb.addResultSet('another')
            nb.addPendingResult(params, '1234')

        # resolve the result, then commit
        with HDF5LabNotebook(self._fn).open() as nb:
            nb.resolvePendingResult(rc, '1234')

        # check everything went to the right place
        with HDF5LabNotebook(self._fn).open() as nb:
            self.assertEqual(nb.currentTag(), 'another')
            self.assertEqual(nb.numberOfResults(), 1)
            self.assertEqual(nb.numberOfPendingResults(), 0)
            nb.select(nb.DEFAULT_RESULTSET)
            self.assertEqual(nb.numberOfResults(), 0)
            self.assertEqual(nb.numberOfPendingResults(), 0)

    def testLockingResultSets(self):
        '''Test that a locked result set stays locked.'''
        e = SampleExperiment()
        nb = HDF5LabNotebook(self._fn, create=True)

        # put some results into the default result set
        params1 = dict(k=10)
        rc1 = e.set(params1).run()
        nb.addResult(rc1)
        params1['k'] = 40
        rc1 = e.set(params1).run()
        nb.addResult(rc1)
        nb.commit()

        # put other results into another set and lock it
        nb.addResultSet('second')
        params2 = dict(k=11)
        rc2 = e.set(params2).run()
        nb.addResult(rc2)
        params3 = dict(k=16)
        nb.addPendingResult(params3, '1234')
        nb.current().finish()
        nb.commit()

        # load the notebook and make sure it's got the correct structure
        nb1 = HDF5LabNotebook(self._fn)
        self.assertEqual(nb1.currentTag(), 'second')
        self.assertTrue(nb1.current().isLocked())
        self.assertEqual(nb1.current().numberOfResults(), 2)
        self.assertEqual(nb1.current().numberOfPendingResults(), 0)
        nb1.select(LabNotebook.DEFAULT_RESULTSET)
        self.assertFalse(nb1.current().isLocked())

    def testLockingNotebook(self):
        '''Test notebook locking.'''
        e = SampleExperiment()
        nb = HDF5LabNotebook(self._fn, create=True)

        rc1 = e.set(dict(k=10)).run()
        rc2 = e.set(dict(k=20)).run()
        rc3 = e.set(dict(k=30)).run()
        rc4 = e.set(dict(k=40)).run()

        # first result set
        nb.addResultSet('first')
        nb.addResult(rc1)
        nb.addResult(rc2)
        self.assertEqual(nb.numberOfResults(), 2)

        # second
        nb.addResultSet('second')
        nb.addResult(rc3)
        nb.addPendingResult(rc4[Experiment.PARAMETERS], '2345')
        self.assertEqual(nb.numberOfResults(), 1)

        # lock the notebook
        nb.finish()

        # check we can't add new result sets
        with self.assertRaises(LabNotebookLockedException):
            nb.addResultSet('third')

        # check the notebook is still locked when reloaded
        nb.commit()
        with HDF5LabNotebook(self._fn).open() as nb1:
            self.assertTrue(nb1.isLocked())
            with self.assertRaises(LabNotebookLockedException):
                nb.addResultSet('third')
            rs = nb1.select('first')
            self.assertTrue(rs.isLocked())
            self.assertEqual(rs.numberOfResults(), 2)
            rs = nb1.select('second')
            self.assertTrue(rs.isLocked())
            self.assertEqual(rs.numberOfPendingResults(), 0)
            self.assertEqual(rs.numberOfResults(), 2)
            rcs = nb1.resultsFor(rc4[Experiment.PARAMETERS])
            self.assertEqual(len(rcs), 1)

    @unittest.skip('Seems to depend on exact config of the underlying filesystem')
    def testNoWriteWhenLocked(self):
        '''Test that no writing happens after a notebook is locked.'''
        nb = HDF5LabNotebook(self._fn, create=True)

        # commit the finished notebook
        nb.addResultSet('first')
        nb.finish()
        t1 = os.path.getmtime(self._fn)

        # reload and commit again, which shouldn't write anything
        with HDF5LabNotebook(self._fn).open() as nb1:
            nb1.commit()
        t2 = os.path.getmtime(self._fn)

        self.assertEqual(t1, t2)

    def testLoadingFromURL(self):
        '''Test we can load a read-only notebook from a URL.'''
        with HDF5LabNotebook(testFileURL).open() as nb1:
            self.assertIn('first', nb1.resultSets())
            self.assertTrue(nb1.isLocked())

    def testRagged(self):
        '''Test we can create ragged axes (i.e., variable length array results).'''
        nb = HDF5LabNotebook(self._fn, create=True)

        params = dict()
        params['k'] = 1
        e = SampleExperiment2()
        rc = e.set(params).run()
        nb.addResult(rc)

        params['k'] = 10
        rc = e.set(params).run()
        nb.addResult(rc)

        self.assertEqual(len(nb.dataframe()), 2)
        nb.commit()

        nb1 = HDF5LabNotebook(self._fn)
        self.assertEqual(len(nb1.dataframe()), 2)
        rcs = nb.resultsFor(dict(k=1))
        self.assertEqual(len(rcs), 1)
        rcs1 = nb1.resultsFor(dict(k=1))
        self.assertEqual(len(rcs1), 1)
        self.assertCountEqual(rcs[0][Experiment.RESULTS]['list'], rcs1[0][Experiment.RESULTS]['list'])
        rcs = nb.resultsFor(dict(k=10))
        self.assertEqual(len(rcs), 1)
        rcs1 = nb1.resultsFor(dict(k=10))
        self.assertEqual(len(rcs1), 1)
        self.assertCountEqual(rcs[0][Experiment.RESULTS]['list'], rcs1[0][Experiment.RESULTS]['list'])

    def testDeleteResultSet(self):
        '''Test a deleted result set is deleted when persisted.'''
        nb = HDF5LabNotebook(self._fn, create=True)

        # create additional result sets and save them
        nb.addResultSet('second')
        params = dict()
        params['k'] = 1
        e = SampleExperiment2()
        rc = e.set(params).run()
        nb.addResult(rc)
        nb.addResultSet('third')
        params = dict()
        params['k'] = 5
        e = SampleExperiment2()
        rc = e.set(params).run()
        nb.addResult(rc)
        nb.commit()

        # delete the result set and re-save
        nb1 = HDF5LabNotebook(self._fn)
        nb1.deleteResultSet('second')
        nb1.commit()

        # make sure we only have the default result set
        nb2 = HDF5LabNotebook(self._fn)
        self.assertEqual(len(nb2), 2)
        self.assertEqual(nb2.currentTag(), 'third')
        self.assertCountEqual(nb2.resultSets(), ['third', LabNotebook.DEFAULT_RESULTSET])

    def testArray(self):
        '''Test we can add arrays as parameters.'''
        params = dict()
        params['k'] = 100

        with HDF5LabNotebook(self._fn, create=True).open() as nb:
            e = SampleExperiment2()
            rc = e.set(params).run()
            nb.addResult(rc)

        with HDF5LabNotebook(self._fn).open() as nb:
            rc = nb.resultsFor(params)[0]
            self.assertCountEqual(rc[Experiment.RESULTS]['list'], [100] * 100)

    def testAllTypes(self):
        '''Test we can store all the types we're likely to be interested in.'''
        rc = Experiment.resultsdict()
        rc[Experiment.METADATA][Experiment.STATUS] = True

        # elementary types
        rc[Experiment.RESULTS]['a_int'] = 10
        rc[Experiment.RESULTS]['b_float'] = 4.5
        rc[Experiment.RESULTS]['c_complex'] = 4.5 + 2j
        rc[Experiment.RESULTS]['d_str'] = 'hello'
        rc[Experiment.RESULTS]['e_bool'] = False
        #rc[Experiment.RESULTS]['f_empty'] = None

        # arrays
        rc[Experiment.RESULTS]['a_int[]'] = [10, 20, 30]
        rc[Experiment.RESULTS]['b_float[]'] = [4.5, 2.0]
        rc[Experiment.RESULTS]['c_complex[]'] = [4.5 + 2j] # of one element
        rc[Experiment.RESULTS]['d_str[]'] = ['hello', 'abandon']
        rc[Experiment.RESULTS]['e_bool[]'] = [False, True, False]
        rc[Experiment.RESULTS]['f_empty[]'] = []

        with HDF5LabNotebook(self._fn, create=True).open() as nb:
            nb.addResult(rc)

        with HDF5LabNotebook(self._fn).open() as nb:
            df = nb.dataframe()
            for k in rc[Experiment.RESULTS].keys():
                if k.find('[]') == -1:
                    # simple type
                    self.assertEqual(rc[Experiment.RESULTS][k], df[k][0])
                else:
                    # array
                    self.assertCountEqual(rc[Experiment.RESULTS][k], df[k][0])


if __name__ == '__main__':
    unittest.main()



# Tests of local-parallel lab class
#
# Copyright (C) 2016--2020 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import *
import unittest
from multiprocessing import cpu_count


class SampleExperiment(Experiment):
    '''A very simple experiment that adds up its parameters.'''

    def do( self, param ):
        total = 0
        for k in param:
            total = total + param[k]
        return dict(total = total)


class SampleExperiment1(SampleExperiment):
    '''An experiment that adds a parameter.'''

    def setUp(self, params):
        super().setUp(params)
        params['additional'] = 10


class ParallelLabTests(unittest.TestCase):

    def testCoresSelection(self):
        '''Test that core selection works as expected.'''

        # default
        self._lab = ParallelLab()
        self.assertEqual(self._lab.numberOfCores(), cpu_count())

        # zero is the same
        self._lab = ParallelLab(cores=0)
        self.assertEqual(self._lab.numberOfCores(), cpu_count())

        # fixed numbers, possibly more than we have physical cores (which
        # will be capped)
        for i in range(1, cpu_count() + 2):
            self._lab = ParallelLab(cores=i)
            self.assertEqual(self._lab.numberOfCores(), min(i, cpu_count()))

    @unittest.skipIf(cpu_count() < 2, 'Need multiple cores to check free core selection')
    def testFreeCoresSelection(self):
        '''Test we can leave cores free.'''
        maxcores = cpu_count()

        # check we leave cores free
        for i in range(1, maxcores - 1):
            self._lab = ParallelLab(cores=-i)    # leave i cores free
            self.assertEqual(self._lab.numberOfCores(), maxcores - i)

        # check we always have at least 1
        self._lab = ParallelLab(cores=-(maxcores + 1))
        self.assertEqual(self._lab.numberOfCores(), 1)

    def testSequential(self):
        '''Test a sequential run.'''
        self._lab = ParallelLab(cores=1)
        self._lab['k'] = range(10)
        self._lab.runExperiment(SampleExperiment())

        # check what we got back
        rcs = self._lab.results()
        self.assertEqual(len(rcs), 10)
        self.assertCountEqual(list(map(lambda rc: rc[Experiment.RESULTS]['total'], rcs)), range(10))

    @unittest.skipIf(cpu_count() < 2, 'Need multiple cores to check parallel execution')
    def testParallel(self):
        '''Test we can run in parallel.'''
        self._lab = ParallelLab()
        self._lab['k'] = range(10)
        self._lab.runExperiment(SampleExperiment())

        # check what we got back
        rcs = self._lab.results()
        self.assertEqual(len(rcs), 10)
        self.assertCountEqual(list(map(lambda rc: rc[Experiment.RESULTS]['total'], rcs)), range(10))

    def testAddParameters(self):
        '''Test that adding experimental parameters works.'''
        self._lab = ParallelLab()
        self._lab['k'] = range(10)
        self._lab.runExperiment(SampleExperiment1())

        # check what we got back
        rcs = self._lab.results()
        self.assertEqual(len(rcs), 10)
        for rc in rcs:
            self.assertIn('additional', rc[Experiment.PARAMETERS])
        self.assertCountEqual(list(map(lambda rc: rc[Experiment.RESULTS]['total'], rcs)),
                              [i + 10 for i in range(10)])


if __name__ == '__main__':
    unittest.main()



# Tests of setting parameters from within experiments
#
# Copyright (C) 2016--2021 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import *
import unittest
import time


class SampleExperiment1(Experiment):
    '''Update parameters from do().'''

    def do(self, params):
        params['k'] = 10
        return dict()


class SampleExperiment2(Experiment):
    '''Update parameters from setUp().'''

    def setUp(self, params):
        super().setUp(params)
        params['k'] = 20


class TestParameterUpdate(unittest.TestCase):

    def testFromDo(self):
        '''Test we can update parameters from do().'''
        params = dict()
        params['a'] = 1
        e = SampleExperiment1()
        rc = e.set(params).run()
        self.assertEqual(rc[Experiment.PARAMETERS]['a'], 1)
        self.assertIn('k', rc[Experiment.PARAMETERS])
        self.assertEqual(rc[Experiment.PARAMETERS]['k'], 10)

    def testFromSetUp(self):
        '''Test we can update parameters from setUp().'''
        params = dict()
        params['a'] = 1
        e = SampleExperiment2()
        rc = e.set(params).run()
        self.assertEqual(rc[Experiment.PARAMETERS]['a'], 1)
        self.assertIn('k', rc[Experiment.PARAMETERS])
        self.assertEqual(rc[Experiment.PARAMETERS]['k'], 20)


if __name__ == '__main__':
    unittest.main()



# Base class for experimental designs
#
# Copyright (C) 2016--2021 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from typing import Dict, List, Tuple, Any
from epyc import Experiment


# Type alias for sets of experiments
ExperimentalConfiguration = List[Tuple[Experiment, Dict[str, Any]]]    #: Type of a configuration of experiments.


class DesignException(Exception):
    '''An exception raised whenever a set of parameter ranges can't
    be used as the basis for a design.

    '''

    def __init__(self, msg: str):
        super().__init__(f'Design excpetion: {msg}')


class Design:
    '''Base class for experimental designs.

    A "design" is a protocol for conducting a set of experiments so as
    to maximise the amount of useful data collected. It is a common
    topic in real-world experiments, and can be applied to
    computational experiments as well.

    A design in ``epyc`` converts a set of :term:`experimental
    parameters` into ann :term:`experimental configuration`, a list
    consisting of pairs of an experiment to run and the parameters at
    which to run it.

    A design must be able to cope with being passed None as an
    experiment, and should return None for all the experiments in the
    configuration: this allows for pre-checks to be performed.

    A design is associated with each :class:`Lab`. By default the standard
    :class:`FactorialDesign` is used, and no further action is needed. Other
    designs can be selected at lab creation time.

    '''

    def experiments(self, e: Experiment, ps: Dict[str, Any]) -> ExperimentalConfiguration:
        '''Convert a mapping from parameter name to list of values into
        a list of mappings from parameter names to single values paired with
        experiment to run at that point, according to the requirements of the
        design. This method must be overridden by sub-classes.

        :param ps: a dict of parameter values
        :returns: an experimental configuration'''
        raise NotImplementedError('parameterSpace() must be overridden by sub-classes')



# Initialisation for epyc package
#
# Copyright (C) 2016--2020 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

"""`epyc` is a Python module for controlling a long-running series of
computational experiments, as is often found when writing simulations
of complex networks and other such domains. There is often a need to
perform a computation across a multi-dimensional parameter space,
varying the parameters, performing and aggregating multiple
repetitions, and wrangling results for analysis and
presentation. Often the experiments being performed are on such a
scale as to require the use of a computing cluster to perform multiple
experiments simultaneously.

Managing all these tasks is complicated, so `epyc` tries to automate
the process. It provides a way to define a :term:`lab` that performs
an :term:`experiment` (or, more likely, a sequence of experiments
according to some :term:experimental design`) whose parameters and
results are recorded as a :term:`result set` in a lab :term:`notebook`
for later retrieval. Laboratories can be sequential (for a single
machine) or parallel (to use a multicore or cluster of machines); lab
notebooks can be persistent to allow experiments to be fired-off and
their results retrieved later -- handy if you use a laptop. Notebooks
store all the data and metadata immutably in a portable format (JSON
or HDF5) to improve the reproducibility of computational experiments.

`epyc` includes a small set of common experimental designs as well as
some :term:`experiment combinators` that separate the logic of a
single experiment from the logic of performing multiple repetitions
and other structuring tasks.

"""

# String written into every persistent notebook file
PackageContactInfo = 'Created by epyc, computational experiment management for Python <https://pypi.org/project/epyc/>'

# Library-specific logger name
Logger = 'epyc'

# Experiments
from .experiment import Experiment, ResultsDict, ExperimentalParameters
from .experimentcombinator import ExperimentCombinator
from .repeatedexperiment import RepeatedExperiment
from .summaryexperiment import SummaryExperiment

# Result sets anf notebooks
from .resultset import ResultSet, ResultSetLockedException, CancelledException, PendingResultException
from .labnotebook import LabNotebook, ResultsStructureException, NotebookVersionException, LabNotebookLockedException
from .jsonlabnotebook import JSONLabNotebook
from .hdf5labnotebook import HDF5LabNotebook

# Experimental designs
from .design import Design, ExperimentalConfiguration, DesignException
from .standard_designs import FactorialDesign, PointwiseDesign

# Labs
from .lab import Lab
from .parallellab import ParallelLab
from .clusterlab import ClusterLab

# Late and/or complex initialisation
Experiment._init_statics()
ResultSet._init_statics()



# Simulation "lab notebook" for collecting results, HDF5 version
#
# Copyright (C) 2020--2021 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

import logging
import os
import h5py                    # type: ignore
from tempfile import NamedTemporaryFile
import numpy                   # type: ignore
from datetime import datetime
from dateutil.parser import parse
from requests import get
from requests.exceptions import MissingSchema
from epyc import Logger, ResultSet, Experiment, LabNotebook, NotebookVersionException, PackageContactInfo
import sys
from typing import Any
if sys.version_info >= (3, 8):
    from typing import Final
else:
    from typing_extensions import Final


logger = logging.getLogger(Logger)


class HDF5LabNotebook(LabNotebook):
    '''A lab notebook that persists itself to an HDF5 file.  `HDF5
    <https://www.hdfgroup.org/>`_ is a very common format for sharing
    large scientific datasets, allowing ``epyc`` to interoperate with
    a larger toolchain.

    ``epyc`` is built on top of the ``h5py`` `Python binding to HDF5
    <https://www.h5py.org/>`_, which handles most of the heavy lifting
    using a lot of machinery for typing and so on matched with
    ``numpy``. Note that the limitations of HDF5's types mean that
    some values may have different types when read than when
    acquired. (See :ref:`hdf5-type-management` for details.)

    The name of the notebook can be a file or a URL. Only files can be
    created or updated: if a URL is provided then the notebook will be
    read and immediately marked as locked. This implies that
    ``create=True`` won't work in conjunction with URLs.

    .. important ::

        Note that because of the design of the ``requests`` library
        used for handling URLs, using a ``file:``-schema URL will
        result in an exception being raised. Use filenames for
        accessing files.

    :param name: HDF5 file or URL backing the notebook
    :param create: (optional) if True, erase any existing file (defaults to False)
    :param description: (optional) free text description of the notebook

    '''

    # Latest file format, defining how a notebook is laid out within the HDF5 file
    Version : Final[str] = '1'                       #: File structure version number used by this notebook.

    # Tuning parameters
    DefaultDatasetSize : int = 10                    #: Default initial size for a new HDF5 dataset.

    # Dataset structure
    RESULTS_DATASET : Final[str] = 'results'         #: Name of results dataset within the HDF5 group for a result set.
    PENDINGRESULTS_DATASET : Final[str] = 'pending'  #: Name of pending results dataset within the HDF5 group for a result set.

    # Metadata for the notebook and/or result sets
    CREATOR : Final[str] = 'creator'                 #: Attribute holding library information and URL.
    VERSION : Final[str] = 'version'                 #: Attribute holding the version of file structure used.
    DESCRIPTION : Final[str] = 'description'         #: Attribute holding the notebook and result set descriptions.
    CURRENT : Final[str] = 'current-resultset'       #: Attribute holding the tag of the current result set.
    LOCKED : Final[str] = 'locked'                   #: Attribute flagging a result set or notebook as being locked to further changes.

    def __init__(self, name : str, create : bool =False, description : str =None):
        # create an empty file structure
        self._file : h5py.File = None          # HDF5 file for underlying data
        self._group : h5py.Group = None        # group associated with the current result set

        # if we're looking at a URL, some behaviour is different
        try:
            # attempt to open the name as a URL
            response = get(name)

            # if we get here, the name is a URL
            self._isRemote = True
        except MissingSchema:
            # name is a file
            self._isRemote = False

        # check for file needing creation
        created = (not self._isRemote) and (create or not os.path.isfile(name))
        if created:
            self._create(name)

        # perform the normal initialisation
        super().__init__(name, description)

        if not created:
            # load notebook from file if it wasn't newly created
            self._open()
            self._load()
            self._close()

            # use the description we were given, if there is one
            if description is not None:
                self.setDescription(description)


    # ---------- Persistence ----------

    def isPersistent(self) -> bool:
        """Return True to indicate the notebook is persisted to an HDF5 file.

        :returns: True"""
        return True

    def commit(self):
        """Persist any changes in the result sets in the notebook to disc."""
        if not self.isLocked():
            self._open()
            self._save()
            self._close()

    def _commit(self):
        '''If we're finishing, commit regardless of the lock status.'''
        self._open()
        self._save()
        self._close()


    # ---------- Managing result sets ----------

    def addResultSet(self, tag : str, description : str =None) -> ResultSet:
        '''Add the necessary structure to the underlying file when creating the
        new result set. This ensures that, even if no results are added,
        there will be structure in the persistent store to indicate that the result
        set was created.

        :param tag: the tag
        :param description: (optional) the description
        :returns: the result set'''
        rs = super(HDF5LabNotebook, self).addResultSet(tag, description)

        # add the appropriate group to the HDF5 file
        self._open()
        if tag not in self._file:
            self._file.create_group(tag)

        # mark the result set as being dirty to force a save
        rs.dirty()

        return rs


    # ---------- Protocol for managing the HDF5 file----------

    def _create(self, name : str):
        '''Create the HDF5 file to back this notebook.

        :param name: the filename
        :param description: the free text description of this notebook'''
        self._file = h5py.File(name, 'w')
        self._file.attrs[self.VERSION] = self.Version
        #self._current = self.resultSet(self.DEFAULT_RESULTSET)
        logger.info(f'Created notebook {name}')

    def _open(self):
        '''Open the HDF5 file that backs this notebook.'''
        if self._file is None:
            if self._isRemote:
                # open as a URL
                response = get(self.name())

                # slurp the contents into a local temp file
                tf = NamedTemporaryFile(delete=False)
                for chunk in response.iter_content(chunk_size=1024):
                    tf.write(chunk)
                tf.close()

                # open the file read-only
                self._file = h5py.File(tf.name, 'r')   # can only be read
            else:
                # open as a file
                self._file = h5py.File(self.name(), 'a')

    def _close(self):
        '''Close the underlying HDF5 file.'''
        if self._file is not None:
            self._file.close()
            self._file = None

        # if we loaded from a URL, lock the notebook
        if self._isRemote:
            self.finish(commit=False)   # don't try to commit it when finishing

    def _purge(self):
        '''Delete any HDF5 datasets that relate to deleted result sets.'''
        tags = self.resultSets()
        g = self._file
        for tag in g.keys():
            if tag not in tags:
                # result set with this tag has been deleted, remove the corresponding dataset
                del g[tag]
                logger.info(f'Purged ResultSet {tag}')

    def _write(self, tag : str):
        '''Write the given result set to the file.

        :tag: the result set tag'''
        rs = self.resultSet(tag)
        names = rs.names()
        g = self._file[tag]

        # delete any attributes
        for k in g.attrs.keys():
            del g.attrs[k]

        # if the result set's type has changed, delete any existing datasets
        if rs.isTypeChanged():
            if self.RESULTS_DATASET in g:
                del g[self.RESULTS_DATASET]
            if self.PENDINGRESULTS_DATASET in g:
                del g[self.PENDINGRESULTS_DATASET]

        # write out the description
        g.attrs.create(self.DESCRIPTION, rs.description(), dtype=h5py.string_dtype())

        # write out all attributes (as strings)
        for k in rs.keys():
            v = rs[k]
            g.attrs.create(k, self._asString(v), dtype=h5py.string_dtype())

        # write out the locked flag
        g.attrs.create(self.LOCKED, rs.isLocked(), dtype=bool)

        if rs.numberOfResults() > 0:
            # ---- PART 1: write structure ---

            # create results table if there isn't one
            if self.RESULTS_DATASET not in g:
                # get the dtype of the result set
                dtype = rs.dtype()
                hdf5dtype = self._HDF5dtype(dtype)

                # create the results dataset
                g.create_dataset(self.RESULTS_DATASET, (self.DefaultDatasetSize,), maxshape=(None,), dtype=hdf5dtype)

                # write out the names of all the fields as attributes
                for d in [ Experiment.METADATA, Experiment.PARAMETERS, Experiment.RESULTS ]:
                    ns = names[d]
                    if ns is not None:
                        g[self.RESULTS_DATASET].attrs.create(d, ns, dtype=h5py.string_dtype())

            # get the HDF5 dataset associated with this result set
            ds = g[self.RESULTS_DATASET]

            # ---- PART 2: write results ---

            # ensure that the dataset is large enough to hold the result set,
            # expanding it if necessary
            if len(ds) != len(rs):
                ds.resize((len(rs),))

            # write out each result
            df = rs.dataframe()
            dtype = rs.dtype()
            hdf5dtype = ds.dtype
            dfnames = dtype.names
            for i in range(len(df.index)):
                # get the row to write
                res = df.iloc[i]

                # convert row in the results table to a line in the dataset
                entry = []
                for k in dfnames:
                    if k in [ Experiment.START_TIME, Experiment.END_TIME ] and isinstance(res[k], datetime):
                        # patch known datestamps ISO-format strings in metadata
                        dt = res[k].isoformat()
                        entry.append(dt)
                    elif k in [Experiment.EXCEPTION]:
                        # patch exception to string
                        entry.append(str(res[k]))
                    else:
                        et = type(res[k])
                        if et is list:
                            # array of something
                            if len(res[k]) > 0 and (type(res[k][0]) == str or type(res[k][0]) == bytes):
                                # string array, convert
                                entry.append(numpy.array(list(map(lambda s: self._asString(s), res[k]))))
                            else:
                                entry.append(numpy.array(res[k]))
                        elif et == str:
                            # string, convert
                            entry.append(self._asString(res[k]))
                        else:
                            # "normal" type, pass through
                            entry.append(res[k])

                # write out the row
                ds[i] = tuple(entry)

        # ---- PART 3: write pending results ---

        if rs.numberOfPendingResults() == 0:
            if self.PENDINGRESULTS_DATASET in g:
                # table isn't needed any more, so delete it to keep things tidy
                del g[self.PENDINGRESULTS_DATASET]
        else:
            pdtype = rs.pendingdtype()
            hdf5pdtype = self._HDF5dtype(pdtype)

            # create results dataset if there isn't one
            if self.PENDINGRESULTS_DATASET not in g:
                # construct the HDF5 type of the pending results
                # create the pending results dataset
                g.create_dataset(self.PENDINGRESULTS_DATASET, (self.DefaultDatasetSize,), maxshape=(None,), dtype=hdf5pdtype)

            # get the pending results dataset
            pds = g[self.PENDINGRESULTS_DATASET]

            # match size to the pending results
            if len(pds) != rs.numberOfPendingResults():
                pds.resize((rs.numberOfPendingResults(),))

            # write out each pending result
            pdf = rs._pending
            if pdf is not None:
                pdfnames = names[Experiment.PARAMETERS]
                for i in range(len(pdf.index)):
                    res = pdf.iloc[i]
                    entry = []
                    for k in pdfnames:
                        if hdf5pdtype[k] == h5py.string_dtype():
                            entry.append(self._asString(res[k]))
                        else:
                            entry.append(res[k])
                    entry.append(res[ResultSet.JOBID])

                    # write out the row
                    pds[i] = tuple(entry)

        # mark the result set as now clean
        rs.dirty(False)
        rs.typechanged(False)

    def _read(self, tag : str):
        '''Read the given result set into memory.

        :param tag: the result set tag'''
        rs = self.resultSet(tag)
        names = dict()
        g = self._file[tag]

        # read the attributes, all treated as strings
        formats = None
        locked = False
        for k in g.attrs.keys():
            if k == self.DESCRIPTION:
                # description held separately from attributes
                rs.setDescription(str(g.attrs[self.DESCRIPTION]))
            elif k == self.LOCKED:
                # locked flag
                locked = g.attrs[self.LOCKED]
            else:
                #rs[k] = self._asString(g.attrs[k])
                rs[k] = str(g.attrs[k])

        if self.RESULTS_DATASET in g:
            # ---- PART 1: read structure ---

            # get the HDF5 dataset associated with this result set
            ds = self._file[tag][self.RESULTS_DATASET]

            # read the names of all the fields and all the attributes
            for k in [ Experiment.METADATA, Experiment.PARAMETERS, Experiment.RESULTS ]:
                if k in ds.attrs:
                    names[k] = list(ds.attrs[k])
                else:
                    names[k] = []   # no result fields (suggests any results so far have failed)

            # ---- PART 2: read results ---

            # read each line of the dataset into the result set
            # sd: this uses the results dict API and so is quite wasteful
            hdf5dtype = ds.dtype
            for i in range(len(ds)):
                entry = list(ds[i])
                rc = Experiment.resultsdict()
                j = 0
                for d in [ Experiment.METADATA, Experiment.PARAMETERS, Experiment.RESULTS ]:
                    for k in names[d]:
                        et = type(entry[j])
                        if et == bytes:
                            # bytes, convert to string
                            entry[j] = entry[j].decode()
                        elif et == numpy.bool_:
                            # internal bool_, convert to ordinary bool
                            entry[j] = bool(entry[j])
                        elif et == numpy.ndarray:
                            # array of something
                            if len(entry[j]) > 0 and type(entry[j][0]) == bytes:
                                # strings, convert
                                entry[j] = list(map(lambda s: s.decode(), list(entry[j])))
                        if k in [ Experiment.START_TIME, Experiment.END_TIME ]:
                            try:
                                # patch known datestamps to datetime objects
                                dt = parse(entry[j])
                                entry[j] = dt
                            except:
                                # leave unchanged if we can't patch
                                pass
                        rc[d][k] = entry[j]
                        j += 1

                self.addResult(rc, tag)

        if self.PENDINGRESULTS_DATASET in g:
            # ---- PART 3: read pending results ---

            # get the pending results dataset
            pds = self._file[tag][self.PENDINGRESULTS_DATASET]

            # extract the parameter names, dropping the job id
            names[Experiment.PARAMETERS] = list(pds.dtype.names)
            names[Experiment.PARAMETERS].remove(ResultSet.JOBID)

            # read in each pending result
            # sd: this uses the results dict API and so is quite wasteful
            pdfnames = list(pds.dtype.names)
            jobidcol = pdfnames.index(ResultSet.JOBID)
            for i in range(len(pds)):
                jobid = ''
                elements = list(pds[i])
                params = dict()
                for j in range(len(pdfnames)):
                    if j == jobidcol:
                        jobid = elements[j]
                    else:
                        params[pdfnames[j]] = elements[j]

                # some backends force strings, not bytes, for job ids
                if isinstance(jobid, bytes):
                    jobid = jobid.decode()

                # record the pending result
                self.addPendingResult(params, jobid, tag)

        # lock the result set if flagged
        if locked:
            rs.finish()

        # mark the result set as clean
        rs.dirty(False)
        rs.typechanged(False)

    def _save(self):
        '''Save all dirty result sets. These are written out completely.'''
        for tag in self.resultSets():
            rs = self.resultSet(tag)
            if rs.isDirty():
                # result set has changed since it was created, loaded, or last saved
                self._write(tag)

        # delete any datasets that aare no longer in the notebook,
        # i.e., those that have been deleted
        self._purge()

        # write out the housekeeping information for the notebook
        meta = self._file.attrs
        meta[self.CREATOR] = PackageContactInfo
        meta[self.VERSION] = self.Version
        meta[self.DESCRIPTION] = self.description()
        meta[self.CURRENT] = self.resultSetTag(self.current())
        meta[self.LOCKED] = self.isLocked()

    def _load(self):
        '''Load the notebook and all result sets.'''
        meta = self._file.attrs

        # version check
        v = meta[self.VERSION]
        if v != self.Version:
            # at present can't handle different file structure versions
            raise NotebookVersionException(self.Version, v)

        # read the tags of all result sets and create empty sets for them
        tags = list(self._file.keys())
        for tag in tags:
            # create an empty result set to hold this dataset
            self.addResultSet(tag)

            # read the result set in
            self._read(tag)

        # read the current result set and select it
        c = meta[self.CURRENT]
        super().select(c)

        # read other attributes
        self._description = meta[self.DESCRIPTION]

        # lock if necessary
        if meta[self.LOCKED]:
            self.finish(commit=False)


    # ---------- Type conversion ----------

    def _HDF5simpledtype(self, dtype : numpy.dtype) -> numpy.dtype:
        '''Patch a simple ``numpy`` dtype to the formats available in HDF5.

        :param dtype: the ``numpy`` dtype
        :returns: the HDF5 dtype'''
        if dtype.kind in ['S', 'U']:
            # h5py can't handle Python strings but has its own dtype for them
            return h5py.string_dtype()
        #elif dtype.kind in ['O']:
        #    # object (probably numpy array), wrap as vlen
        #    return h5py.vlen_dtype(dtype)
        elif dtype.shape != ():
            # list, make into a vlen type
            return h5py.vlen_dtype(self._HDF5dtype(dtype.base))
        else:
            # leave unpatched
            return dtype

    def _HDF5dtype(self, dtype : numpy.dtype) -> numpy.dtype:
        '''Patch a ``numpy`` dtype into its HDF5 equivalent. This method
        handles structured types with named fields.

        :param dtype: the ``numpy`` dtype
        :returns: the HDF5 dtype'''
        if dtype.names is not None:
            # a structured dtype, unpack and patch all the fields
            elements = []
            for k in dtype.names:
                t = self._HDF5dtype(dtype.fields[k][0])
                elements.append((k, t))
            return numpy.dtype(elements)
        else:
            # a simple dtype, patch it
            return self._HDF5simpledtype(dtype)

    def _asString(self, o : Any) -> str:
        '''Coerce the given value to a string. This is needed to handle
        occasional (and Python-version-dependent) weirdness in the
        representation of strings as raw sequences when going to and from
        HDF5.

        :param o: the object
        :returns: the string'''
        if isinstance(o, bytes):
            return o.decode(encoding='ascii')
        elif isinstance(o, str):
            #return o
            return o.encode(encoding='ascii')
        else:
            #return str(o)
            return str(o).encode(encoding='ascii')



# Container script for manipulating notebooks from the command line
#
# Copyright (C) 2021 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

import os
import sys
import re
import click
import epyc

notebook_re = re.compile('^([\w/\.-]+)$')
notebook_resultset_re = re.compile('^([\w/\.-]+)?:([\w/\.-]+)$')
notebook_opt_resultset_re = re.compile('^([\w/\.-]+)(:([\w/\.-]+))?$')
notebook_resultset_rename_re = re.compile('^([\w/\.-]+)?:([\w/\.-]+)(=([\w/\.-]+))?$')

@click.group()
def cli():
    '''A command line interface (CLI) to epyc lab notebooks.'''
    pass

@cli.command()
@click.argument('notebook')
@click.option('-l/-s', '--long/--short', default=True, help='long/short form output')
def show(notebook, long):
    '''Show the structure of a notebook.

    The long form displays a human-readable summary of the result sets
    and their descriptions. The short form simply lists the result
    sets, one per line, in a form suitable for feeding to other
    commands. The default is the long form. The current result set is
    starred.

    '''

    # open the notebook
    try:
        nb = epyc.HDF5LabNotebook(name=notebook, create=False)
    except Exception as e:
        print(f"Can't open {notebook}: {e}", file=sys.stderr)
        sys.exit(1)

    # display the notebook's structure, in long or short form
    if long:
        # long form, print metadata about notebook
        desc=nb.description()
        click.echo(f'{desc}' +
                   (click.style(' (locked)', fg='red') if nb.isLocked() else ''))
        click.echo()

        # iterate the result sets
        currentTag = nb.currentTag()
        click.echo('Result sets:')
        for tag in nb.resultSets():
            # show result set global summary
            rs = nb.resultSet(tag)
            n = len(rs)
            desc = rs.description()
            click.echo(click.style(f'  {tag}', fg='green') +
                       (click.style('(*)', fg='yellow') if tag == currentTag else '') +
                       f' [{n}]: {desc}' +
                       (click.style(' (locked)', fg='red') if nb.isLocked() else ''))

            # show any attributes
            for key in rs.keys():
                value = rs[key]
                click.echo(click.style(f'    {key}', fg='green') + f': {value}')
    else:
        # short form, just print result set tags
        print('\n'.join(nb.resultSets()))

@cli.command()
@click.argument('rss', nargs=-1)
@click.argument('dest', nargs=1)
@click.option('-v', '--verbose', count=True, help='Generate verbose output (repeat for extra verbosity)')
@click.option('-n', '--pretend', is_flag=True, help="Check validity but don't copy anything")
def copy(rss, dest, verbose, pretend):
    '''Copy one or more result sets to the destination notebook.

    Result sets are specified as a triple [NOTEBOOK]:TAG[=NEWTAG]
    where NOTEBOOK is the name of a notebook, TAG is a result set tag
    within NOTEBOOK, and NEWTAG is a tag for the result set when it's
    copied to the destination (to avoid name clashes). If NOTEBOOK is
    omitted then the same notebook as the previous result set is
    used. If NEWTAG is omitted then TAG is used for the destination
    result set as well

    '''

    # check we have result sets to copy
    if len(rss) == 0:
        print('No result sets to copy')
        exit(0)

    # iterate through all result set specifiers
    copies = []
    fn = None
    tag = None
    newtag = None
    for spec in rss:
        # extract the parts from the specifier
        m = notebook_resultset_rename_re.match(spec)
        if m is None:
            print(f"Invalid result set specifier '{spec}'", file=sys.stderr)
            exit(1)
        if m[1] == '':
            # no notebook, can we use the previous one?
            if fn is None:
                # no notebook default
                print(f"No notebook for result set specifier '{spec}'", file=sys.stderr)
                exit(1)
        else:
            # notebook replaces the current one
            fn = m[1]
            tag = m[2]
        if m[4] is None:
            # no rename, use the same tag
            newtag = tag
        else:
            # result set will be renamed when copied
            newtag = m[4]

        # save the decomposed specifier
        copies.extend([(fn, tag, newtag)])

    # open the destination
    with epyc.HDF5LabNotebook(name=dest).open() as nb:
        # check that destination is not locked
        if nb.isLocked():
            print(f'Destination notebook {dest} is locked')
            sys.exit(1)

        # traverse the copy specifiers
        ofn = None
        nb1 = None
        for (fn, tag, newtag) in copies:
            # change notebooks if needed
            if ofn != fn:
                if nb1 is not None:
                    # commit the previous notebook
                    nb1.commit()

                # load the notebook if it's not already loaded
                nb1 = epyc.HDF5LabNotebook(fn)
            ofn = fn

            # sanity check the result sets
            if tag not in nb1.resultSets():
                print(f"No result set '{tag}' in {fn}", file=sys.stderr)
                sys.exit(1)
            if newtag in nb.resultSets():
                print(f"Result set '{newtag}' already exists in {dest}", file=sys.stderr)
                sys.exit(1)

            if verbose > 0:
                click.echo('Copied ' +
                           click.style(f'{fn}:{tag}', fg='green') +
                           ' -> ' +
                           click.style(f'{dest}:{newtag}', fg='green'))
            if not pretend:
                # copy the result set
                rs1 = nb1.resultSet(tag)
                rs = nb.addResultSet(newtag)
                for rc in rs1.results():
                    rs.addSingleResult(rc)

@cli.command()
@click.argument('rss', nargs=-1)
@click.option('-v', '--verbose', count=True, help='Generate verbose output (repeat for extra verbosity)')
@click.option('-n', '--pretend', is_flag=True, help="Check validity but don't copy anything")
def remove(rss, verbose, pretend):
    '''Remove one or more result sets from notebook(s).

    Result sets are specified as a pair [NOTEBOOK]:TAG where
    NOTEBOOK is the name of a notebook and TAG is a result set tag
    within NOTEBOOK. If NOTEBOOK is omitted then the same notebook as
    the previous result set is used.

    '''

    # iterate through all result set specifiers
    copies = []
    fn = None
    tag = None
    for spec in rss:
        # extract the parts from the specifier
        m = notebook_resultset_re.match(spec)
        if m is None:
            print(f"Invalid result set specifier '{spec}'", file=sys.stderr)
            exit(1)
        if m[1] == '':
            # no notebook, can we use the previous one?
            if fn is None:
                # no notebook default
                print(f"No notebook for result set specifier '{spec}'", file=sys.stderr)
                exit(1)
        else:
            # notebook replaces the current one
            fn = m[1]
            tag = m[2]
        copies.extend([(fn, tag)])

    # traverse the removal specifiers
    ofn = None
    nb1 = None
    for (fn, tag) in copies:
        # change notebooks if needed
        if ofn != fn:
            if nb1 is not None:
                # commit the previous notebook
                nb1.commit()

            # load the notebook if it's not already loaded
            nb1 = epyc.HDF5LabNotebook(fn)
        ofn = fn

        # sanity check the result sets
        if tag not in nb1.resultSets():
            print(f"No result set '{tag}' in {fn}", file=sys.stderr)
            sys.exit(1)

        if verbose > 0:
            click.echo('Removed ' +
                       click.style(f'{fn}:{tag}', fg='green'))
        if not pretend:
            nb1.deleteResultSet(tag)

@cli.command()
@click.argument('spec')
@click.option('-v', '--verbose', count=True, help='Generate verbose output (repeat for extra verbosity)')
@click.option('-n', '--pretend', is_flag=True, help="Check validity but don't change anything")
def select(spec, verbose, pretend):
    '''Select a result set as current.

    The result set is specified as a pair NOTEBOOK[:TAG] where
    NOTEBOOK is the name of a notebook and TAG is a result set tag
    within NOTEBOOK. If no TAG is specified then the command returns
    the tag of the current result set.

    '''

    # extract the notebook and tag
    m = notebook_opt_resultset_re.match(spec)
    if m is None:
        print(f"Invalid result set specifier '{spec}'", file=sys.stderr)
        exit(1)
    else:
        fn = m[1]
        tag = m[3]

    with epyc.HDF5LabNotebook(fn).open() as nb:
        if tag == None:
            # no tag, just display the current result set's tag
            tag = nb.currentTag()
            print(tag)
        else:
            # sanity check
            if tag in nb.resultSets():
                # tag exists, make it current
                if not pretend:
                    nb.select(tag)
                    if verbose > 0:
                        click.echo('Made ' +
                                   click.style(f'{fn}:{tag}', fg='green') +
                                   " current")
            else:
                print(f'No result set {tag}')
                exit(1)


if __name__ == '__main__':
    cli()



# Tests of standard experimental designs
#
# Copyright (C) 2016--2021 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import *
import unittest


class SampleExperiment(Experiment):

    def do(self, params):
        return dict()


class TestParameterUpdate(unittest.TestCase):

    def setUp(self):
        self._e = SampleExperiment()


    # ---------- General ----------

    def testOverriding(self):
        '''Test we can't call on the base class.'''
        d = Design()
        with self.assertRaises(NotImplementedError):
            d.experiments(self._e, dict)


    # ---------- Factorial design ----------

    def testEmptyFactorial(self):
        '''Test empty factorial design.'''
        params = dict()
        es = FactorialDesign().experiments(self._e, params)
        self.assertCountEqual(es, [])

    def testSingleParameterFactorial(self):
        '''Test single-parameter factorial design.'''
        params = dict(a=range(3))
        es = FactorialDesign().experiments(self._e, params)
        self.assertCountEqual(es, [(self._e, dict(a=0)), (self._e, dict(a=1)), (self._e, dict(a=2))])

    def testTwoParameterFactorial(self):
        '''Test two-parameter factorial design.'''
        params = dict(a=range(3),
                      b=[1, 2])
        es = FactorialDesign().experiments(self._e, params)
        self.assertCountEqual(es, [(self._e, dict(a=0, b=1)), (self._e, dict(a=1, b=1)), (self._e, dict(a=2, b=1)),
                                   (self._e, dict(a=0, b=2)), (self._e, dict(a=1, b=2)), (self._e, dict(a=2, b=2))])

    def testTwoParameterFactorialOneEmpty(self):
        '''Test two-parameter factorial design, one of which is empty.'''
        params = dict(a=range(3),
                      b=[])
        es = FactorialDesign().experiments(self._e, params)
        self.assertCountEqual(es, [(self._e, dict(a=0)), (self._e, dict(a=1)), (self._e, dict(a=2))])

    def testTwoParameterFactorialOnePointwise(self):
        '''Test two-parameter factorial design, one of which is a pointwise value.'''
        params = dict(a=range(3),
                      b=[6])
        es = FactorialDesign().experiments(self._e, params)
        self.assertCountEqual(es, [(self._e, dict(a=0, b=6)), (self._e, dict(a=1, b=6)), (self._e, dict(a=2, b=6))])


    # ---------- Pointwise design ----------

    def testEmptyPointwise(self):
        '''Test empty pointwise design.'''
        params = dict()
        es = PointwiseDesign().experiments(self._e, params)
        self.assertCountEqual(es, [])

    def testSingleParameterPointwise(self):
        '''Test single-parameter pointwise design.'''
        params = dict(a=range(3))
        es = PointwiseDesign().experiments(self._e, params)
        self.assertCountEqual(es, [(self._e, dict(a=0)), (self._e, dict(a=1)), (self._e, dict(a=2))])

    def testTwoParameterPointwise(self):
        '''Test two-parameter pointwise design.'''
        params = dict(a=range(3),
                      b=[4, 5, 6])
        es = PointwiseDesign().experiments(self._e, params)
        self.assertCountEqual(es, [(self._e, dict(a=0, b=4)), (self._e, dict(a=1, b=5)), (self._e, dict(a=2, b=6))])

    def testTwoParameterPointwiseUnequal(self):
        '''Test two-parameter pointwise design with unequal ranges.'''
        params = dict(a=range(3),
                      b=[4, 5])
        with self.assertRaises(DesignException):
            es = PointwiseDesign().experiments(self._e, params)

    def testTwoParameterPointwiseOneExtending(self):
        '''Test two-parameter pointwise design where one parameter is a singleton value.'''
        params = dict(a=range(3),
                      b=[4])
        es = PointwiseDesign().experiments(self._e, params)
        self.assertCountEqual(es, [(self._e, dict(a=0, b=4)), (self._e, dict(a=1, b=4)), (self._e, dict(a=2, b=4))])

    def testThreeParameterPointwiseTwoExtending(self):
        '''Test two-parameter pointwise design where two parameters are singleton values.'''
        params = dict(a=range(3),
                      b=[4],
                      c=[8])
        es = PointwiseDesign().experiments(self._e, params)
        self.assertCountEqual(es, [(self._e, dict(a=0, b=4, c=8)), (self._e, dict(a=1, b=4, c=8)), (self._e, dict(a=2, b=4, c=8))])

    def testThreeParameterPointwiseOneExtending(self):
        '''Test two-parameter pointwise design where two parameters are singleton values.'''
        params = dict(a=range(3),
                      b=[5, 6, 7],
                      c=[8])
        es = PointwiseDesign().experiments(self._e, params)
        self.assertCountEqual(es, [(self._e, dict(a=0, b=5, c=8)), (self._e, dict(a=1, b=6, c=8)), (self._e, dict(a=2, b=7, c=8))])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

# Setup for epyc
#
# Copyright (C) 2016--2022 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from setuptools import setup  # type: ignore

with open('README.rst') as f:
    longDescription = f.read()

setup(name='epyc',
      version='1.7.1',
      description='Python computational experiment management',
      long_description=longDescription,
      url='http://github.com/simoninireland/epyc',
      author='Simon Dobson',
      author_email='simoninireland@gmail.com',
      license='License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Science/Research',
                   'Intended Audience :: Developers',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7',
                   'Programming Language :: Python :: 3.8',
                   'Programming Language :: Python :: 3.9',
                   'Programming Language :: Python :: 3.10',
                   'Topic :: Scientific/Engineering'],
      python_requires='>=3.5',
      packages=['epyc',
                'epyc.scripts',
                ],
      include_package_data=True,
      package_data={'epyc': ['py.typed']},
      entry_points='''
        [console_scripts]
        epyc=epyc.scripts.epyc:cli
      ''',
      zip_safe=False,
      install_requires=["numpy >= 1.17.5", "pyzmq", "ipyparallel >= 6.2.4", "cloudpickle", "pandas", "h5py", "joblib", "requests", "click", ],
      extra_requires={':python_version < 3.8': ['typing_extensions']})



# Tests of sequential lab class
#
# Copyright (C) 2016--2020 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

import unittest
from datetime import datetime
import numpy
from epyc import *


class SampleExperiment(Experiment):
    '''A very simple experiment that adds up its parameters.'''

    def do( self, param ):
        total = 0
        for k in param:
            total = total + param[k]
        return dict(total = total)


class NullExperiment(Experiment):
    '''An experiment that just returns the value of its (one)
    parameter as its result.'''

    def do( self, param ):
        k = list(param.keys())[0]
        return dict(result = param[k])


class LabTests(unittest.TestCase):

    def setUp( self ):
        '''Create a lab in which to perform tests.'''
        self._lab = Lab()

    def testDefaultDesign(self):
        '''Test the default design is a factorial design.'''
        self.assertTrue(isinstance(self._lab.design(), FactorialDesign))

    def testParameterOne( self ):
        '''Test that we can set a single parameter.'''
        self._lab['a'] = numpy.arange(0, 100)

        self.assertEqual(len(self._lab['a']), len(numpy.arange(0, 100)))
        for i in numpy.arange(0, 100):
            self.assertIn(i, self._lab['a'])

    def testParameterThree( self ):
        '''Test that we can set several parameters.'''
        self._lab['a'] = numpy.arange(0, 100)
        self._lab['b'] = numpy.arange(0, 1000)
        self._lab['c'] = numpy.arange(10, 50, 10)

        self.assertCountEqual(self._lab['a'], numpy.arange(0, 100))
        self.assertCountEqual(self._lab['b'], numpy.arange(0, 1000))
        self.assertCountEqual(self._lab['c'], numpy.arange(10, 50, 10))

    def testRunOne( self ):
        '''Test that a simple experiment runs against a 1D parameter space.'''
        n = 100

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment())
        res = self._lab.results()

        # check that the whole parameter space has a result
        self.assertEqual(len(res), n)
        for p in res:
            self.assertIn(p[Experiment.PARAMETERS]['a'], r)

        # check that each result corresponds to its parameter
        for p in res:
            self.assertEqual(p[Experiment.PARAMETERS]['a'], p[Experiment.RESULTS]['total'])

    def testRunTwo( self ):
        '''Test that a simple experiment runs against a 2D parameter space.'''
        n = 10

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab['b'] = r
        self._lab.runExperiment(SampleExperiment())
        res = self._lab.results()

        # check that the whole parameter space has a result
        self.assertEqual(len(res), n * n)
        for p in res:
            self.assertIn(p[Experiment.PARAMETERS]['a'], r)
            self.assertIn(p[Experiment.PARAMETERS]['b'], r)

        # check that each result corresponds to its parameter
        for p in res:
            self.assertEqual(p[Experiment.PARAMETERS]['a'] + p[Experiment.PARAMETERS]['b'], p[Experiment.RESULTS]['total'])

    def testReady(self):
        '''Test we can check readiness correctly.'''
        n = 10

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab['b'] = r
        self._lab.runExperiment(SampleExperiment())
        self.assertTrue(self._lab.ready())
        self.assertEqual(self._lab.readyFraction(), 1.0)

    def testSinglePoint( self ):
        '''Test that using a single point as a range still works.'''
        n = 100

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab['b'] = 0
        self._lab.runExperiment(SampleExperiment())
        res = self._lab.results()

        # check that the whole parameter space has a result
        self.assertEqual(len(res), n)
        for p in res:
            self.assertIn(p[Experiment.PARAMETERS]['a'], r)
            self.assertEqual(p[Experiment.PARAMETERS]['b'], 0)

        # check that each result corresponds to its parameter
        for p in res:
            self.assertEqual(p[Experiment.PARAMETERS]['a'] + p[Experiment.PARAMETERS]['b'], p[Experiment.RESULTS]['total'])

    def testStringsNotUnpacked( self ):
        '''Test we don't unpack strings, even though they're iterable'''
        v = 'A string'

        self._lab['s'] = v
        self._lab.runExperiment(NullExperiment())
        res = self._lab.results()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0][Experiment.RESULTS]['result'], v)

    def testContains(self):
        '''Test the contains operator.'''
        self._lab['a'] = 10
        self._lab['b'] = range(10)
        self.assertIn('a', self._lab)
        self.assertIn('b', self._lab)
        del self._lab['a']
        self.assertNotIn('a', self._lab)
        self.assertIn('b', self._lab)

    def testDeletion(self):
        '''Test we can delete parameters.'''
        self.assertCountEqual(self._lab.parameters(), [])

        # delete one
        self._lab['a'] = 10
        self._lab['b'] = range(10)
        self.assertCountEqual(self._lab.parameters(), ['a', 'b'])
        del self._lab['a']
        self.assertCountEqual(self._lab.parameters(), ['b'])

        # delete all
        self._lab['c'] = range(10)
        self._lab.deleteAllParameters()
        self.assertCountEqual(self._lab.parameters(), [])

    def testLength(self):
        '''Test lengths of the various designs.'''

        # under a factorial design
        self._lab = Lab(design=FactorialDesign())
        self._lab['a'] = range(10)
        self._lab['b'] = range(20)
        self.assertEqual(len(self._lab), len(self._lab['a']) * len(self._lab['b']))

        # under a pointwise design
        self._lab = Lab(design=PointwiseDesign())
        self._lab['a'] = range(10)
        self._lab['b'] = range(10)
        self.assertEqual(len(self._lab), len(self._lab['a']))


    # ---------- createWith ----------

    def _resultsdict(self):
        '''Set up a results dict populated with plausible metadata.'''
        _rc = Experiment.resultsdict()
        _rc[Experiment.METADATA][Experiment.EXPERIMENT] = str(self.__class__)
        _rc[Experiment.METADATA][Experiment.START_TIME] = datetime.now()
        _rc[Experiment.METADATA][Experiment.END_TIME] = datetime.now()
        _rc[Experiment.METADATA][Experiment.SETUP_TIME] = 10
        _rc[Experiment.METADATA][Experiment.EXPERIMENT_TIME] = 20
        _rc[Experiment.METADATA][Experiment.TEARDOWN_TIME] = 10
        _rc[Experiment.METADATA][Experiment.ELAPSED_TIME] = 40
        _rc[Experiment.METADATA][Experiment.STATUS] = True
        return _rc

    def create(self, lab):
        '''Create some results into the current result set.'''
        rc1 = self._resultsdict()
        rc1[Experiment.PARAMETERS]['a'] = 1
        rc1[Experiment.PARAMETERS]['b'] = 2
        rc1[Experiment.RESULTS]['first'] = 3
        self._lab.notebook().addResult(rc1)

    def createFail(self, lab):
        '''Fail to create results.'''
        raise Exception('Failed!')

    def createEmpty(self, lab):
        '''Test we have no parameters in the lab.'''
        self.assertEqual(len(lab.parameters()), 0)

    def createMess(self, lab):
        '''Delete the result set we were expecting to revert to.'''
        lab.notebook().deleteResultSet('one')
        raise Exception('Failed (after creating mess)!')

    def testBasicCreateWith(self):
        '''Test createWith works.'''
        rc = self._lab.createWith('test1', self.create, description='A first set')
        self.assertTrue(rc)
        self.assertEqual(self._lab.notebook().current().description(), 'A first set')
        self.assertIn('test1', self._lab.notebook().resultSets())
        self.assertFalse(self._lab.notebook().current().isLocked())

    def testCreateWithFinishes(self):
        '''Test createWith finishes a result set when requested.'''
        rc = self._lab.createWith('test1', self.create, finish=True)
        self.assertTrue(rc)
        self.assertTrue(self._lab.notebook().current().isLocked())

    def testCreateWithFail(self):
        '''Test createWith handles deletion on failure.'''
        with self.assertRaises(Exception):
            rc = self._lab.createWith('test1', self.createFail)
        self.assertNotIn('test1', self._lab.notebook().resultSets())

    def testCreateWithFailNoDelete(self):
        '''Test createWith can be prevented from deleting on failure.'''
        with self.assertRaises(Exception):
            rc = self._lab.createWith('test1', self.createFail, delete=False)
        self.assertIn('test1', self._lab.notebook().resultSets())
        self.assertFalse(self._lab.notebook().current().isLocked())

    def testCreateWithFailNoDeleteNotFinished(self):
        '''Test createWith doesn't lock a partial result set.'''
        with self.assertRaises(Exception):
            rc = self._lab.createWith('test1', self.createFail,
                                      delete=False, finish=True)
        self.assertIn('test1', self._lab.notebook().resultSets())
        self.assertFalse(self._lab.notebook().current().isLocked())

    def testCreateWithNoPropagate(self):
        '''Test createWith hides the exception on failure.'''
        rc = self._lab.createWith('test1', self.createFail, propagate=False)
        self.assertFalse(rc)
        self.assertNotIn('test1', self._lab.notebook().resultSets())

    def testClearParameterSpace(self):
        '''Test the parameter space is blanked before calling the creation function.'''
        self._lab['a'] = 10
        rc = self._lab.createWith('test1', self.createEmpty)

    def testDontClearParameterSpace(self):
        '''Test the parameter space blanking can be inhibited.'''
        self._lab['a'] = 10
        with self.assertRaises(Exception):
            rc = self._lab.createWith('test1', self.createEmpty,
                                      deleteAllParameters=False)

    def testRevertResultSetOnFailure(self):
        '''Test we revert to the correct result set on failure.'''
        self._lab.notebook().addResultSet('one')
        self._lab.notebook().addResultSet('two')
        self._lab.notebook().select('one')
        with self.assertRaises(Exception):
            rc = self._lab.createWith('test1', self.createFail)
        self.assertEqual(self._lab.notebook().currentTag(), 'one')
        self.assertCountEqual(self._lab.notebook().resultSets(),
                              ['one', 'two', LabNotebook.DEFAULT_RESULTSET])

    def testMess(self):
        '''Test we deal with reversion problems.'''
        self._lab.notebook().addResultSet('one')
        self.assertCountEqual(self._lab.notebook().resultSets(),
                              ['one', LabNotebook.DEFAULT_RESULTSET])
        with self.assertRaises(Exception):
            rc = self._lab.createWith('test1', self.createMess)
        self.assertEqual(self._lab.notebook().currentTag(), LabNotebook.DEFAULT_RESULTSET)
        self.assertCountEqual(self._lab.notebook().resultSets(),
                              [LabNotebook.DEFAULT_RESULTSET])


if __name__ == '__main__':
    unittest.main()



# Tests of JSON notebooks
#
# Copyright (C) 2016--2020 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import *

import unittest
import os
from tempfile import NamedTemporaryFile


class SampleExperiment(Experiment):
    '''A very simple experiment that adds up its parameters.'''

    def do( self, param ):
        total = 0
        for k in param:
            total = total + param[k]
        return dict(total = total)


class SampleExperiment1(Experiment):
    '''An experiment that fails.'''

    def do( self, param ):
        raise Exception('A (deliberate) failure')


class JSONLabNotebookTests(unittest.TestCase):

    def setUp( self ):
        '''Set up with a temporary file.'''
        tf = NamedTemporaryFile()
        tf.close()
        self._fn = tf.name
        #self._fn = 'test.json'

    def tearDown( self ):
        '''Delete the temporary file.'''
        try:
            os.remove(self._fn)
            #pass
        except OSError:
            pass

    def testCreate( self ):
        '''Test creation of empty notebook (which won't create file)'''
        js = JSONLabNotebook(self._fn, description="A test notebook", create=True)
        self.assertTrue(js.isPersistent())
        self.assertEqual(js.name(), self._fn)

    def testCreateAndSave( self ):
        '''Test creation and saving of notebook'''
        e = SampleExperiment()
        params1 = dict( a = 1, b = 2 )
        rc1 = e.set(params1).run()
        params2 = dict( a = 1, b = 3 )
        rc2 = e.set(params2).run()

        js = JSONLabNotebook(self._fn, description="A test notebook", create=True)
        js.addResult(rc1)
        js.addPendingResult(params2, '2')
        js.commit()

        jsr = JSONLabNotebook(self._fn)
        self.assertEqual(jsr.description(), "A test notebook")
        self.assertCountEqual(jsr.pendingResults(), js.pendingResults())
        self.assertCountEqual(jsr.results(), js.results())

    def testCreateAndUpdate( self ):
        '''Test creation and updating of notebook'''
        e = SampleExperiment()
        params1 = dict( a = 1, b = 2 )
        rc1 = e.set(params1).run()
        params2 = dict( a = 1, b = 3 )
        rc2 = e.set(params2).run()

        js = JSONLabNotebook(self._fn, description="A test notebook", create=True)
        js.addResult(rc1)
        js.addPendingResult(params2, '2')
        js.commit()

        js.resolvePendingResult(rc2, '2')
        js.commit()

        jsr = JSONLabNotebook(self._fn)
        self.assertEqual(jsr.description(), "A test notebook")
        self.assertEqual(len(jsr.pendingResults()), 0)
        self.assertCountEqual(jsr.results(), js.results())

    def testCreateOverwrite( self ):
        '''Test the create flag'''
        e = SampleExperiment()
        params1 = dict( a = 1, b = 2 )
        rc1 = e.set(params1).run()

        js = JSONLabNotebook(self._fn, description="A test notebook", create=True)
        js.addResult(rc1)
        js.commit()

        jsr = JSONLabNotebook(self._fn, description="A test notebook", create=True)
        self.assertEqual(jsr.description(), "A test notebook")
        self.assertEqual(len(jsr.results()), 0)
        self.assertEqual(len(jsr.pendingResults()), 0)

    def testCreateNoOverwrite( self ):
        '''Test that the create flag being false doesn't overwrite'''
        e = SampleExperiment()
        params1 = dict( a = 1, b = 2 )
        rc1 = e.set(params1).run()

        js = JSONLabNotebook(self._fn, description="A test notebook", create=True)
        js.addResult(rc1)
        js.commit()

        # check we keep the results but change the description
        jsr = JSONLabNotebook(self._fn, description = "Nothing to see")
        self.assertEqual(jsr.description(), "Nothing to see")
        self.assertCountEqual(jsr.results(), js.results())
        self.assertEqual(len(jsr.pendingResults()), 0)

    def testReadEmpty( self ):
        '''Test we can correctly load an empty file, resulting in an empty notebook'''
        e = SampleExperiment()
        params1 = dict( a = 1, b = 2 )
        rc1 = e.set(params1).run()

        # notebook with a result
        js = JSONLabNotebook(self._fn, description="A test notebook", create=True)
        js.addResult(rc1)
        js.commit()

        # ...which we then wipe out by re-creating it
        js2 = JSONLabNotebook(self._fn, create=True,
                              description="Another test notebook")

        # ...and check we did
        jsr = JSONLabNotebook(self._fn)
        self.assertEqual(len(jsr.results()), 0)
        self.assertEqual(len(jsr.pendingResults()), 0)

        # ...and that it commits and reloads properly
        js2.commit()
        jsr = JSONLabNotebook(self._fn)
        self.assertEqual(len(jsr.results()), 0)
        self.assertEqual(len(jsr.pendingResults()), 0)

    def testTimePatching( self ):
        '''Check that the timing metadata is persisted correctly.'''
        e = SampleExperiment()
        params1 = dict( a = 1, b = 2 )
        rc1 = e.set(params1).run()

        js = JSONLabNotebook(self._fn, description="A test notebook", create=True)
        js.addResult(rc1)
        js.commit()

        js2 = JSONLabNotebook(self._fn)
        rc2 = (js2.results())[0]
        self.assertEqual(rc1[Experiment.METADATA][Experiment.START_TIME],
                         rc2[Experiment.METADATA][Experiment.START_TIME])
        self.assertEqual(rc1[Experiment.METADATA][Experiment.END_TIME],
                         rc2[Experiment.METADATA][Experiment.END_TIME])


    def testPersistingException( self ):
        '''Test we persist exceptions as strings.'''
        e = SampleExperiment1()
        params1 = dict(a = 1)
        rc1 = e.set(params1).run()

        js = JSONLabNotebook(self._fn, description="A test notebook", create=True)
        js.addResult(rc1)
        js.commit()

        js2 = JSONLabNotebook(self._fn)
        rc2 = (js2.results())[0]
        self.assertFalse(rc2[Experiment.METADATA][Experiment.STATUS])
        self.assertTrue(isinstance(rc2[Experiment.METADATA][Experiment.EXCEPTION], str))

    def testMultipleResultSets(self):
        '''Check we keep results sets separate.'''
        e = SampleExperiment()
        js = JSONLabNotebook(self._fn, create=True)

        # put some results into the default result set
        params1 = dict(a=10, b=20)
        rc1 = e.set(params1).run()
        js.addResult(rc1)
        params1['b'] = 40
        rc1 = e.set(params1).run()
        js.addResult(rc1)

        # put more results, with different type, into a different result set
        js.addResultSet('second')
        params2 = dict(c=10, d=20, e=90)
        rc2 = e.set(params2).run()
        js.addResult(rc2)
        params2['d'] = 40
        js.addPendingResult(params2, '123')
        js.commit()

        # check we've maintained the correct current result set
        js = JSONLabNotebook(self._fn)
        self.assertEqual(js.currentTag(), 'second')

        # check we've kept the results separate
        self.assertEqual(js.numberOfResults(), 1)
        self.assertEqual(js.numberOfPendingResults(), 1)
        self.assertEqual(len(js.resultsFor(dict(d=20))), 1)
        js.select(LabNotebook.DEFAULT_RESULTSET)
        self.assertEqual(js.numberOfResults(), 2)
        self.assertEqual(js.numberOfPendingResults(), 0)
        self.assertEqual(len(js.resultsFor(dict(a=10))), 2)
        self.assertEqual(len(js.resultsFor(dict(b=20))), 1)

    def testLocking(self):
        '''Test that a locked result set stays locked.'''
        e = SampleExperiment()
        js = JSONLabNotebook(self._fn, create=True)

        # put some results into the default result set
        params1 = dict(a=10, b=20)
        rc1 = e.set(params1).run()
        js.addResult(rc1)
        params1['b'] = 40
        rc1 = e.set(params1).run()
        js.addResult(rc1)
        js.commit()

        # put other results into another set and lock it
        js.addResultSet('second')
        params2 = dict(a=11, b=22)
        rc2 = e.set(params2).run()
        js.addResult(rc2)
        params3 = dict(a=11, b=24)
        js.addPendingResult(params3, '1234')
        js.current().finish()
        js.commit()

        # load the notebook and makle sure it's got the correct structure
        js1 = JSONLabNotebook(self._fn)
        self.assertEqual(js1.currentTag(), 'second')
        self.assertTrue(js1.current().isLocked())
        self.assertEqual(js1.current().numberOfResults(), 2)
        self.assertEqual(js1.current().numberOfPendingResults(), 0)
        js1.select(LabNotebook.DEFAULT_RESULTSET)
        self.assertFalse(js1.current().isLocked())

    def testLockingNotebook(self):
        '''Test notebook locking.'''
        e = SampleExperiment()
        nb = JSONLabNotebook(self._fn, create=True)

        rc1 = e.set(dict(k=10)).run()
        rc2 = e.set(dict(k=20)).run()
        rc3 = e.set(dict(k=30)).run()
        rc4 = e.set(dict(k=40)).run()

        # first result set
        nb.addResultSet('first')
        nb.addResult(rc1)
        nb.addResult(rc2)

        # second
        nb.addResultSet('second')
        nb.addResult(rc3)
        nb.addPendingResult(rc4[Experiment.PARAMETERS], '2345')

        # lock the notebook
        nb.finish()

        # check we can't add new result sets
        with self.assertRaises(LabNotebookLockedException):
            nb.addResultSet('third')

        # check the notebook in still  locked when reloaded
        nb.commit()
        with JSONLabNotebook(self._fn).open() as nb1:
            self.assertTrue(nb1.isLocked())
            with self.assertRaises(LabNotebookLockedException):
                nb.addResultSet('third')
            rs = nb1.select('first')
            self.assertTrue(rs.isLocked())
            self.assertEqual(rs.numberOfResults(), 2)
            rs = nb1.select('second')
            self.assertTrue(rs.isLocked())
            self.assertEqual(rs.numberOfPendingResults(), 0)
            self.assertEqual(rs.numberOfResults(), 2)
            rcs = nb1.resultsFor(rc4[Experiment.PARAMETERS])
            self.assertEqual(len(rcs), 1)

    def testDeleteResultSet(self):
        '''Test a deleted result set is deleted when persisted.'''
        nb = JSONLabNotebook(self._fn, create=True)

        # create additional result sets and save them
        nb.addResultSet('second')
        params = dict()
        params['b'] =  1
        e = SampleExperiment()
        rc = e.set(params).run()
        nb.addResult(rc)
        nb.addResultSet('third')
        params = dict()
        params['b'] =  5
        e = SampleExperiment()
        rc = e.set(params).run()
        nb.addResult(rc)
        nb.commit()

        # delete the result set and re-save
        nb1 = JSONLabNotebook(self._fn)
        nb1.deleteResultSet('second')
        nb1.commit()

        # make sure we only have the default result set
        nb2 = JSONLabNotebook(self._fn)
        self.assertEqual(len(nb2), 2)
        self.assertEqual(nb2.currentTag(), 'third')
        self.assertCountEqual(nb2.resultSets(), ['third', LabNotebook.DEFAULT_RESULTSET])

if __name__ == '__main__':
    unittest.main()



# Experiment combinator to run the same experiment several times.
#
# Copyright (C) 2016--2022 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import ExperimentCombinator, Experiment, ResultsDict
import sys
from typing import Any, Dict, List, Union
if sys.version_info >= (3, 8):
    from typing import Final
else:
    # backwards compatibility with Python 35, Python36, and Python37
    from typing_extensions import Final


class RepeatedExperiment(ExperimentCombinator):
    """A experiment combinator that takes a "base" experiment and runs it
    several times. This means you can define a single experiment separate
    from its repeating logic.

    When run, a repeated experiment runs a number of repetitions of the underlying
    experiment at the same point in the parameter space. The result of the
    repeated experiment is the list of results from the underlying experiment.
    If the underlying experiment itself returns a list of results, these are all
    flattened into a single list.

    :param ex: the underlying experiment
    :pamam N: the number of repetitions to perform"""

    # Extra metadata
    REPETITIONS: Final[str] = 'epyc.experiment.repetitions'   #: Metadata element for number of repetitions performed.
    I          : Final[str] = 'epyc.experiment.rep'           #: Metadata element for index in the repetitions.


    def __init__(self, ex: Experiment, N: int ):
        super().__init__(ex)
        self._N = N

    def repetitions(self) -> int:
        """Return the number of repetitions of the underlying experiment
        we expect to perform.

        :returns: the number of repetitions"""
        return self._N

    def do(self, params: Dict[str, Any]) -> List[ResultsDict]:
        """Perform the number of repetitions we want. The results returned
        will be a list of the results dicts generated by the repeated experiments.
        The metadata for each experiment will include an entry
        :attr:`RepeatedExperiment.REPETITIONS` for the number of
        repetitions that occurred (which will be the length of this list) and
        an entry :attr:`RepeatedExperiment.I` for the index of the result in that
        sequence.

        :param params: the parameters to the experiment
        :returns: a list of result dicts"""
        N = self.repetitions()
        e = self.experiment()
        results = []
        for i in range(N):
            rcs = e.run()

            # make sure we have a list to traverse
            if not isinstance(rcs, list):
                rcs = [rcs]

            # add repetition information to each results dict's metadata
            for j in range(len(rcs)):
                rc = rcs[j]
                rc[Experiment.METADATA][self.I] = i
                rc[Experiment.METADATA][self.REPETITIONS] = N
            results.extend(rcs)
        return results



# Standard experimental designs
#
# Copyright (C) 2016--2022 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from typing import Dict, List, Tuple, Any
import numpy
from epyc import Design, DesignException, Experiment, ExperimentalConfiguration


class FactorialDesign(Design):
    '''A simple factorial design.

    In a factorial design, an experiment is perform for every
    combination of a lab's parameters. Essentially this forms the
    cross-product of all parameter values, returned as a list of
    dicts. If the lab was set up with the following parameters:

    .. code-block:: python

       lab['a'] = [1, 2]
       lab['b'] = [3, 4]

    then this design would generate a space consisting of four points:

    - {a=1, b=3}
    - {a=1, b=4}
    - {a=2, b=3}
    - {a=2, b=4}

    at which it would run the given experiment. The experiments are
    returned in random order.

    '''

    def experiments(self, e: Experiment, ps: Dict[str, Any]) -> ExperimentalConfiguration:
        '''Form the cross-product of all parameters.

        :param ps: a dict of parameter values
        :returns: an experimental configuration'''
        ds = []
        for p in ps.keys():
            rs = ps[p]
            dsprime = []
            for r in rs:
                if len(ds) > 0:
                    for (_, dp) in ds:
                        eprime = dp.copy()
                        eprime[p] = r
                        dsprime.append((e, eprime))
                else:
                    eprime = {}
                    eprime[p] = r
                    dsprime.append((e, eprime))
            if len(dsprime) > 0:
                ds = dsprime

        # randomise the order of the experiments
        numpy.random.shuffle(ds)

        return ds


class PointwiseDesign(Design):
    '''A design whose space is the sequence of values taken from the range
    of each parameter. If the lab was set up with the following parameters:

    .. code-block:: python

       lab['a'] = [1, 2]
       lab['b'] = [3, 4]

    then this design would generate a space consisting of two points:

    - {a=1, b=3}
    - {a=2, b=4}

    This design requires that all parameters have the same length of range:
    if a parameter is a singleton (only a single value), this will be extended
    across all the space. So if the parameters were:

    .. code-block:: python

       lab['a'] = 1
       lab['b'] = [3, 4]

    the design would generate:

    - {a=1, b=3}
    - {a=1, b=4}
    '''

    def experiments(self, e: Experiment, ps: Dict[str, Any]) -> ExperimentalConfiguration:
        '''Form experimental points from corresponding values in
        the parameter ranges, extending any singletons.

        :param ps: a dict of parameter values
        :returns: an experimental configuration'''

        # check parameter ranges are all equal or 1
        ls = list(map(len, list(ps.values())))
        sls = set(ls)
        nls = len(sls)
        l = 0
        if nls == 1:
            # only one parameter range length
            l = ls[0]
        elif nls > 1:
            # more than one length, are there too many altogether?
            if nls > 2:
                # too many different ranges
                raise DesignException('Parameter range lengths don\'t match')
            elif 1 not in sls:
                # not extending
                raise DesignException('Parameter range lengths don\'t match an not extending')
            else:
                # some ranges are being extended, extract the other length
                for j in sls:
                    if j != 1:
                        l = j

        # extract the parameters being extended
        ones = [p for (p, v) in ps.items() if len(v) == 1]

        ds = []
        for i in range(l):
            eprime = {}
            for p in ps.keys():
                if p in ones:
                    eprime[p] = ps[p][0]
                else:
                    eprime[p] = ps[p][i]
            ds.append((e, eprime))
        return ds
